# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Views for the django-ca app.

.. seealso::

   * https://docs.djangoproject.com/en/dev/topics/class-based-views/
   * https://django-ca.readthedocs.io/en/latest/python/views.html
"""

import abc
import logging
import secrets
import typing
from datetime import datetime, timezone as tz
from http import HTTPStatus
from typing import Dict, Generic, Iterable, List, Optional, Set, Type, TypeVar, Union, cast

import acme.jws
import josepy as jose
from acme import messages

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509.oid import NameOID

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import transaction
from django.http import Http404, HttpRequest, HttpResponse, JsonResponse
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View

from django_ca import ca_settings
from django_ca.acme.errors import AcmeBadCSR, AcmeException, AcmeForbidden, AcmeMalformed, AcmeUnauthorized
from django_ca.acme.messages import CertificateRequest, NewOrder
from django_ca.acme.responses import (
    AcmeResponse,
    AcmeResponseAccount,
    AcmeResponseAccountCreated,
    AcmeResponseAuthorization,
    AcmeResponseBadNonce,
    AcmeResponseChallenge,
    AcmeResponseError,
    AcmeResponseMalformed,
    AcmeResponseMalformedPayload,
    AcmeResponseNotFound,
    AcmeResponseOrder,
    AcmeResponseOrderCreated,
    AcmeResponseUnauthorized,
    AcmeResponseUnsupportedMediaType,
)
from django_ca.acme.utils import parse_acme_csr
from django_ca.constants import REASON_CODES
from django_ca.models import (
    AcmeAccount,
    AcmeAuthorization,
    AcmeCertificate,
    AcmeChallenge,
    AcmeOrder,
    Certificate,
    CertificateAuthority,
)
from django_ca.tasks import acme_issue_certificate, acme_validate_challenge, run_task
from django_ca.utils import check_name, int_to_hex, validate_email

log = logging.getLogger(__name__)
MessageTypeVar = TypeVar("MessageTypeVar", bound=jose.json_util.JSONObjectWithFields)
DirectoryMetaAlias = Dict[str, Union[str, List[str]]]


if typing.TYPE_CHECKING:
    from django.http.response import HttpResponseBase


class ContactValidationMixin:
    """Mixin for validating contact information."""

    def validate_contacts(self, message: messages.Registration) -> None:
        """Validate the contact information for this message."""

        for contact in message.contact:
            if contact.startswith(messages.Registration.email_prefix):
                addr = contact[len(messages.Registration.email_prefix) :]

                # RFC 8555, section 7.3
                #
                #   Clients MUST NOT provide a "mailto" URL in the "contact" field that contains "hfields"
                #   [RFC6068] or more than one "addr-spec" in the "to" component.

                # We rule our quoted local address fields, otherwise it's extremely hard to validate
                # email addresses.
                if addr.startswith('"'):
                    raise AcmeMalformed("invalidContact", "Quoted local part in email is not allowed.")

                # Since the local part is not quoted, it cannot contain a ',' either, so a ',' means there
                # is more than one "addr-spec" in the "to" component. (see RFC 8555 quote above)
                if "," in addr:
                    raise AcmeMalformed("invalidContact", "More than one addr-spec is not allowed.")

                # Validate that there are no hfields in the address.
                # NOTE: ',' appears to be valid in the local part according to RFC 5322
                _local, domain = addr.split("@", 1)
                if "?" in domain:
                    raise AcmeMalformed("invalidContact", f"{domain}: hfields are not allowed.")

                # Finally, verify that we're getting at least a valid domain.
                try:
                    validate_email(addr)
                except ValueError as ex:
                    raise AcmeMalformed("invalidContact", f"{domain}: Not a valid email address.") from ex
            else:
                # RFC 8555, section 7.3
                #
                #   If the server rejects a contact URL for using an unsupported scheme, it MUST raise an
                #   error of type "unsupportedContact", ...
                raise AcmeMalformed("unsupportedContact", f"{contact}: Unsupported address scheme.")


class AcmeDirectory(View):
    """
    `Equivalent LE URL <https://acme-v02.api.letsencrypt.org/directory>`__

    .. seealso:: `RFC 8555, section 7.1.1 <https://tools.ietf.org/html/rfc8555#section-7.1.1>`_
    """

    def _url(self, request: HttpRequest, name: str, ca: CertificateAuthority) -> str:
        return request.build_absolute_uri(reverse(f"django_ca:{name}", kwargs={"serial": ca.serial}))

    def get(self, request: HttpRequest, serial: Optional[str] = None) -> HttpResponse:
        # pylint: disable=missing-function-docstring; standard Django view function
        if not ca_settings.CA_ENABLE_ACME:
            raise Http404("Page not found.")

        if serial is None:
            try:
                # NOTE: default() already calls usable()
                ca = CertificateAuthority.objects.acme().default()
            except ImproperlyConfigured:
                return AcmeResponseNotFound(message="No (usable) default CA configured.")
        else:
            try:
                # NOTE: Serial is already sanitized by URL converter
                ca = CertificateAuthority.objects.acme().usable().get(serial=serial)
            except CertificateAuthority.DoesNotExist:
                return AcmeResponseNotFound(message=f"{serial}: CA not found.")

        # Get some random data into the directory view, as explained in the Let's Encrypt directory:
        #   https://community.letsencrypt.org/t/adding-random-entries-to-the-directory/33417
        rnd = jose.json_util.encode_b64jose(secrets.token_bytes(16))

        directory: Dict[str, Union[str, DirectoryMetaAlias]] = {
            rnd: "https://community.letsencrypt.org/t/adding-random-entries-to-the-directory/33417",
            "keyChange": "http://localhost:8000/django_ca/acme/todo/key-change",
            "newAccount": self._url(request, "acme-new-account", ca),
            "newNonce": self._url(request, "acme-new-nonce", ca),
            "newOrder": self._url(request, "acme-new-order", ca),
            "revokeCert": self._url(request, "acme-revoke", ca),
        }

        # Construct a "meta" object if and add it if any fields are defined. Note that the meta object is
        # optional (RFC 8555, section 7.1.1: "The object MAY additionally contain a "meta" field.").
        meta: DirectoryMetaAlias = {}
        if ca.website:
            meta["website"] = ca.website
        if ca.terms_of_service:
            meta["termsOfService"] = ca.terms_of_service
        if ca.caa_identity:
            meta["caaIdentities"] = [ca.caa_identity]  # array of string
        if meta:
            directory["meta"] = meta

        return JsonResponse(directory)


class AcmeGetNonceViewMixin:
    """View mixin that provides methods to get and validate a Nonce.

    Note that this mixin depends on the presence of a ``serial`` argument to the URL resolver.
    """

    kwargs: Dict[str, str]
    nonce_length = 32
    """Length of generated Nonces."""

    def get_cache_key(self, nonce: str) -> str:
        """Get the cache key for the given request and nonce."""
        return f"acme-nonce-{self.kwargs['serial']}-{nonce}"

    def get_nonce(self) -> str:
        """Get a random Nonce and add it to the cache."""

        data = secrets.token_bytes(self.nonce_length)
        nonce = jose.json_util.encode_b64jose(data)
        cache.set(self.get_cache_key(nonce), 0)
        return nonce

    def validate_nonce(self, nonce: str) -> bool:
        """Validate that the given nonce was issued and was not used before."""
        try:
            count = cache.incr(self.get_cache_key(nonce))
        except ValueError:
            # raised if cache_key is not set
            return False

        if count > 1:  # nonce was already used
            # NOTE: "incr" returns the *new* value, so "1" is the expected value.
            return False

        return True


@method_decorator(csrf_exempt, name="dispatch")
class AcmeBaseView(AcmeGetNonceViewMixin, View, metaclass=abc.ABCMeta):
    """Base class for all ACME views."""

    requires_key = False  # True if we require a full key (-> new accounts)
    accepts_kid_or_jwk = False  # Set to true to accept both KID or JWK
    jwk: jose.jwk.JWK
    jws: acme.jws.JWS

    @abc.abstractmethod
    def process_acme_request(self, slug: Optional[str]) -> AcmeResponse:
        """Abstract method expected to implement processing a message.

        The `slug` argument is the URL slug that identifies an ACME object and is None for requests that
        either create an object or do not process any object.
        """

    def is_account_usable(self, account: AcmeAccount) -> bool:
        """Method determining if an account is usable.

        Implementing views may override this function if what makes an account "usable" is different. Notably,
        the AccountView considers accounts usable even if the terms of service where not agreed, as the user
        can then agree to the terms of service.
        """
        # Check for various conditions that make an account unusable and raise an appropriate exception
        if account.status == AcmeAccount.STATUS_DEACTIVATED:
            # NOTE: Account may have other status that makes it unusable, these cases are covered by
            # account.usable below.
            raise AcmeUnauthorized(message="Account has been deactivated.")
        if account.ca.terms_of_service and not account.terms_of_service_agreed:
            raise AcmeUnauthorized(message="Account did not agree to the terms of service.")

        # COVERAGE NOTE: The check for the CA is already done when selecting the CA, so account.ca.usable
        # should always be true at this point. Check is left here as an additional precaution.
        if not account.ca.usable:  # pragma: no cover
            raise AcmeUnauthorized(message="Certificate Authority is not usable.")

        # This should return True unless the account was revoked by the server, in which case we let the main
        # function return a generic error message instead.
        return account.usable

    def set_link_relations(self, response: "HttpResponseBase", **kwargs: str) -> None:
        """Set Link relations headers according to RFC8288.

        `RFC8555, section 7.1 <https://tools.ietf.org/html/rfc8555#section-7.1>`_ states:

            The "index" link relation is present on all resources other than the directory and indicates the
            URL of the directory.

        .. seealso:: https://tools.ietf.org/html/rfc8288
        """

        kwargs["index"] = reverse("django_ca:acme-directory", kwargs={"serial": self.kwargs["serial"]})
        response["Link"] = ", ".join(
            f'<{self.request.build_absolute_uri(v)}>;rel="{k}"' for k, v in kwargs.items()
        )

    # def log_request(self):
    #    """Function for logging prepared requests."""
    #    if getattr(settings, 'LOG_ACME_REQUESTS', None):
    #        prepared_key = self.__class__.__name__
    #        prepared_path = os.path.join(settings.FIXTURES_DIR, 'prepared-acme-requests.json')
    #        prepared_data = {}
    #        if os.path.exists(prepared_path):
    #            with open(prepared_path) as stream:
    #                prepared_data = json.load(stream)
    #        if prepared_key not in prepared_data:
    #            prepared_data[prepared_key] = []
    #        prepared_data[prepared_key].append(self.prepared)
    #        with open(prepared_path, 'w') as stream:
    #            json.dump(prepared_data, stream, indent=4)

    # NOINSPECTION/TYPE NOTE: It's okay to be more specific here
    # noinspection PyMethodOverriding
    def dispatch(  # type: ignore[override]
        self, request: HttpRequest, serial: str, slug: Optional[str] = None
    ) -> "HttpResponseBase":
        if not ca_settings.CA_ENABLE_ACME:
            raise Http404("Page not found.")

        # COVERAGE NOTE: Checking just for safety here.
        if not isinstance(serial, str):  # pragma: no cover
            raise ImproperlyConfigured("View expects a str for a serial")
        if slug is not None and not isinstance(slug, str):  # pragma: no cover
            raise ImproperlyConfigured("View expects a str for a slug")

        try:
            response = super().dispatch(request, serial=serial, slug=slug)
        except AcmeException as ex:
            response = ex.get_response()
        except Exception as ex:  # pylint: disable=broad-except
            log.exception(ex)
            response = AcmeResponseError(message="Internal server error")

        self.set_link_relations(response)

        # self.log_request()

        # RFC 8555, section 6.7:
        # An ACME server provides nonces to clients using the HTTP Replay-Nonce header field, as specified in
        # Section 6.5.1.  The server MUST include a Replay-Nonce header field in every successful response to
        # a POST request and SHOULD provide it in error responses as well.
        response["replay-nonce"] = self.get_nonce()
        return response

    def post(self, request: HttpRequest, serial: str, slug: Optional[str] = None) -> AcmeResponse:
        # pylint: disable=missing-function-docstring; standard Django view function
        # pylint: disable=attribute-defined-outside-init
        # pylint: disable=too-many-return-statements; b/c of the many checks

        # TODO: RFC 8555, 6.2 has a nice list of things to check here that we don't yet fully cover
        if request.content_type != "application/jose+json":
            # RFC 8555, 6.2:
            # "Because client requests in ACME carry JWS objects in the Flattened JSON Serialization, they
            # must have the Content-Type header field set to "application/jose+json".  If a request does not
            # meet this requirement, then the server MUST return a response with status code 415 (Unsupported
            # Media Type)."
            return AcmeResponseUnsupportedMediaType()

        # self.prepared['body'] = json.loads(request.body.decode('utf-8'))

        try:
            self.jws = acme.jws.JWS.json_loads(request.body)
        except (jose.errors.DeserializationError, TypeError):
            return AcmeResponseMalformed(message="Could not parse JWS token.")

        combined = self.jws.signature.combined
        if combined.jwk and combined.kid:
            # 'The "jwk" and "kid" fields are mutually exclusive.  Servers MUST reject requests that contain
            # both.'
            return AcmeResponseMalformed(message="jwk and kid are mutually exclusive.")

        # Get certificate authority for this request
        try:
            self.ca = CertificateAuthority.objects.acme().usable().get(serial=serial)
        except CertificateAuthority.DoesNotExist:
            return AcmeResponseNotFound(message="The requested CA cannot be found.")

        if combined.jwk:
            if not self.requires_key and not self.accepts_kid_or_jwk:
                return AcmeResponseMalformed(message="Request requires a JWK key ID.")

            self.jwk = combined.jwk  # set JWK from request
        elif combined.kid:
            if self.requires_key and not self.accepts_kid_or_jwk:
                return AcmeResponseMalformed(message="Request requires a full JWK key.")

            # combined.kid is a full URL pointing to the account.
            try:
                account = AcmeAccount.objects.viewable().get(ca=self.ca, kid=combined.kid)
            except AcmeAccount.DoesNotExist:
                return AcmeResponseUnauthorized(message="Account not found.")

            if self.is_account_usable(account) is False:
                # RFC 855, 7.3.6:
                #
                #   If a server receives a POST or POST-as-GET from a deactivated account, it MUST return an
                #   error response with status code 401 (Unauthorized) and type
                #   "urn:ietf:params:acme:error:unauthorized".
                return AcmeResponseUnauthorized(message="Account not usable.")
            # self.prepared['thumbprint'] = account.thumbprint
            # self.prepared['pem'] = account.pem
            # self.prepared['account_pk'] = account.pk

            self.jwk = jose.jwk.JWK.load(account.pem.encode("utf-8"))  # load JWK from database
            self.account = account
        else:
            # ... 'Either "jwk" (JSON Web Key) or "kid" (Key ID)'
            return AcmeResponseMalformed(message="JWS contained neither key nor key ID.")

        if len(self.jws.signatures) != 1:  # pragma: no cover
            # RFC 8555, 6.2: "The JWS MUST NOT have multiple signatures"
            return AcmeResponseMalformed(message="Multiple JWS signatures encountered.")

        # "The JWS Protected Header MUST include the following fields:..."
        if not combined.alg:  # pragma: no cover
            # ... "alg"
            return AcmeResponseMalformed(message="No algorithm specified.")

        # Verify JWS signature
        try:
            if not self.jws.verify(self.jwk):
                return AcmeResponseMalformed(message="JWS signature invalid.")
        except Exception:  # pylint: disable=broad-except
            return AcmeResponseMalformed(message="JWS signature invalid.")

        # self.prepared['nonce'] = jose.encode_b64jose(combined.nonce)
        if combined.nonce is None or not self.validate_nonce(jose.json_util.encode_b64jose(combined.nonce)):
            # ... "nonce"
            return AcmeResponseBadNonce()

        if combined.url != request.build_absolute_uri():
            # ... "url"
            # RFC 8555 is not really clear on the required response code, but merely says "If the two do not
            # match, then the server MUST reject the request as unauthorized."
            return AcmeResponseUnauthorized(message="URL does not match.")

        return self.process_acme_request(slug=slug)


class AcmePostAsGetView(AcmeBaseView, metaclass=abc.ABCMeta):
    """Base class for ACME post-as-get requests."""

    ignore_body = False  # True if we want to ignore the message body

    @abc.abstractmethod
    def acme_request(self, slug: str) -> AcmeResponse:
        """Abstract method to process an ACME post-as-get request.

        Actual view subclasses are expected to implement this function.

        Note that the `slug` argument is never ``None`` for post-as-get requests, as the request would then
        contain no information.
        """

    def process_acme_request(self, slug: Optional[str]) -> AcmeResponse:
        if self.ignore_body is False and self.jws.payload != b"":
            return AcmeResponseMalformed(message="Non-empty payload in get-as-post request.")
        if slug is None:  # pragma: no cover; just a safety measure
            return AcmeResponseError(message="PostAsGet view called with slug.")

        return self.acme_request(slug=slug)


class AcmeMessageBaseView(AcmeBaseView, Generic[MessageTypeVar], metaclass=abc.ABCMeta):
    """Base class for ACME requests with a message payload."""

    message_cls: Type[MessageTypeVar]

    @abc.abstractmethod
    def acme_request(self, message: MessageTypeVar, slug: Optional[str]) -> AcmeResponse:
        """Process ACME request.

        Actual view subclasses are expected to implement this function.
        """

    def process_acme_request(self, slug: Optional[str]) -> AcmeResponse:
        try:
            message = self.message_cls.json_loads(self.jws.payload)
            log.debug("ACME message: %s", message)
        except jose.errors.DeserializationError as e:
            return AcmeResponseMalformedPayload(message=", ".join(e.args))

        return self.acme_request(message, slug)


class AcmeNewNonceView(AcmeGetNonceViewMixin, View):
    """View to retrieve a new nonce for replay protection.

    Equivalent LE URL: https://acme-v02.api.letsencrypt.org/acme/new-nonce

    .. seealso::

       * `RFC 8555, section 6.5 <https://tools.ietf.org/html/rfc8555#section-6.5>`_
       * `RFC 8555, section 7.2 <https://tools.ietf.org/html/rfc8555#section-7.2>`_
    """

    # NOINSPECTION/TYPE NOTE: It's okay to be more specific here
    # noinspection PyMethodOverriding
    def dispatch(self, request: HttpRequest, serial: str) -> "HttpResponseBase":  # type: ignore[override]
        if not ca_settings.CA_ENABLE_ACME:
            raise Http404("Page not found.")

        # COVERAGE NOTE: Checking just for safety here.
        if not isinstance(serial, str):  # pragma: no cover
            raise ImproperlyConfigured("View expects a str for a serial")

        response = super().dispatch(request, serial)
        response["replay-nonce"] = self.get_nonce()

        # RFC 8555, section 7.2:
        #
        #   The server MUST include a Cache-Control header field with the "no-store" directive in responses
        response["cache-control"] = "no-store"
        return response

    def head(self, request: HttpRequest, serial: str) -> HttpResponse:
        """Get a new Nonce with a HEAD request."""
        # pylint: disable=method-hidden; false positive - View.setup() sets head as property if not defined
        # pylint: disable=unused-argument; false positive - really used by AcmeGetNonceViewMixin
        return HttpResponse()

    def get(self, request: HttpRequest, serial: str) -> HttpResponse:
        """Get a new Nonce with a GET request.

        Note that certbot always does a HEAD request, but RFC 8555, section 7.2 mandates support for GET
        requests.
        """
        # pylint: disable=unused-argument; false positive - really used by AcmeGetNonceViewMixin
        return HttpResponse(status=HTTPStatus.NO_CONTENT)  # 204, unlike HEAD, which has 200


class AcmeNewAccountView(ContactValidationMixin, AcmeMessageBaseView[messages.Registration]):
    """Implements endpoint for creating a new account, that is ``/acme/new-account``.

    This view is called when the ACME client tries to register a new account.

    .. seealso:: `RFC 8555, 7.3 <https://tools.ietf.org/html/rfc8555#section-7.3>`_
    """

    message_cls = messages.Registration
    requires_key = True

    def acme_request(self, message: messages.Registration, slug: Optional[str]) -> AcmeResponseAccount:
        """Process ACME request."""
        pem = (
            self.jwk.key.public_bytes(
                encoding=Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
            )
            .decode("utf-8")
            .strip()
        )
        thumbprint = jose.json_util.encode_b64jose(self.jwk.thumbprint())

        # RFC 8555, section 7.3:
        #
        #   If this field is present with the value "true", then the server MUST NOT create a new account if
        #   one does not already exist.  This allows a client to look up an account URL based on an account
        #   key (see Section 7.3.1).
        if message.only_return_existing:
            try:
                account = AcmeAccount.objects.get(thumbprint=thumbprint, pem=pem)
                return AcmeResponseAccount(self.request, account)
            except AcmeAccount.DoesNotExist as ex:
                # RFC 8555, section 7.3:
                #
                #   ... account does not exist, then the server MUST return an error response with status code
                #   400 (Bad Request) and type "urn:ietf:params:acme:error:accountDoesNotExist".
                raise AcmeMalformed(typ="accountDoesNotExist", message="Account does not exist.") from ex

        # RFC 8555, section 7.3.1
        #
        #   If the server receives a newAccount request signed with a key for which it already has an account
        #   registered with the provided account key, then it MUST return a response with status code 200 (OK)
        #   and provide the URL of that account in the Location header field.
        try:
            # NOTE: Filter for thumbprint too b/c index for the field should speed up lookups.
            account = AcmeAccount.objects.get(ca=self.ca, thumbprint=thumbprint, pem=pem)
            return AcmeResponseAccount(self.request, account)
        except AcmeAccount.DoesNotExist:
            pass

        if self.ca.acme_registration is False:
            raise AcmeUnauthorized(message="Account registration is disabled.")

        if self.ca.acme_requires_contact and not message.emails:
            # NOTE: RFC 8555 does not specify an error code in this case
            raise AcmeUnauthorized(message="Must provide at least one contact address.")

        # Make sure that contact addresses are valid
        self.validate_contacts(message)

        # certbot/acme 1.31.0 does not send this value at all the CA does not have any terms of service.
        # This issue was not present in 1.29.0 and is no longer present in 2.0.0, but any other client may of
        # course not send this value.
        if message.terms_of_service_agreed is None:
            terms_of_service_agreed = False
        else:
            terms_of_service_agreed = message.terms_of_service_agreed

        account = AcmeAccount(
            ca=self.ca,
            contact="\n".join(message.contact),
            status=AcmeAccount.STATUS_VALID,
            terms_of_service_agreed=terms_of_service_agreed,
            thumbprint=thumbprint,
            pem=pem,
        )
        account.set_kid(self.request)

        # Call full_clean() so that model validation can do its magic
        try:
            account.full_clean()
            account.save()
        except ValidationError as ex:
            # Add a pretty list of validation errors to the detail field in the response
            subproblems = ", ".join(
                sorted([f"{k}: {v1.rstrip('.')}" for k, v in ex.message_dict.items() for v1 in v])
            )
            raise AcmeMalformed(message=f"Invalid account: {subproblems}.") from ex

        # self.prepared['thumbprint'] = account.thumbprint
        # self.prepared['pem'] = account.pem
        # self.prepared['account_pk'] = account.pem

        # RFC 8555, section 7.3
        #
        #   The server returns this account object in a 201 (Created) response, with the account URL in a
        #   Location header field.
        #
        # AcmeResponseAccountCreated adds the Location field currently.
        return AcmeResponseAccountCreated(self.request, account)


class AcmeAccountView(ContactValidationMixin, AcmeMessageBaseView[messages.Registration]):
    """View allowing the update of accounts or receiving current account details."""

    message_cls = messages.Registration

    def is_account_usable(self, account: AcmeAccount) -> bool:
        """Overridden to make accounts usable that have not agreed to the terms of service.

        A client may agree to the terms of service after account creation, then making the account usable for
        other operations.
        """
        if account.status == AcmeAccount.STATUS_DEACTIVATED:
            # NOTE: Account may have other status that makes it unusable, these cases are covered by
            # the check if the status is valid below
            raise AcmeUnauthorized(message="Account has been deactivated.")

        # COVERAGE NOTE: The check for the CA is already done when selecting the CA, so account.ca.usable
        # should always be true at this point. Check is left here as an additional precaution.
        if not account.ca.usable:  # pragma: no cover
            raise AcmeUnauthorized(message="Certificate Authority is not usable.")

        return account.status == AcmeAccount.STATUS_VALID

    @transaction.atomic
    def acme_request(self, message: messages.Registration, slug: Optional[str] = None) -> AcmeResponseAccount:
        account = AcmeAccount.objects.get(slug=slug)

        if message.status == AcmeAccount.STATUS_DEACTIVATED:
            # RFC 8555, section 7.3.6 - Account Deactivation
            log.info("Deactivating account %s", account.slug)
            account.status = AcmeAccount.STATUS_DEACTIVATED
            account.save()

            # Cancel all pending operations
            account.orders.filter(status=AcmeOrder.STATUS_PENDING).update(status=AcmeOrder.STATUS_INVALID)
            AcmeAuthorization.objects.filter(
                order__account=account, status=AcmeAuthorization.STATUS_PENDING
            ).update(status=AcmeAuthorization.STATUS_DEACTIVATED)
        elif message.contact:
            self.validate_contacts(message)
            account.contact = "\n".join(message.contact)
            account.save()
        elif message.terms_of_service_agreed is not None:
            account.terms_of_service_agreed = message.terms_of_service_agreed
            account.save()
        else:
            raise AcmeMalformed(message="Only contact information can be updated.")

        return AcmeResponseAccount(self.request, account)


class AcmeAccountOrdersView(AcmeBaseView):
    """View showing orders for an account (not yet implemented)"""

    # TODO: implement this view
    def process_acme_request(self, slug: Optional[str]) -> AcmeResponse:  # pragma: no cover
        raise AcmeException(message="Not Implemented.")


class AcmeNewOrderView(AcmeMessageBaseView[NewOrder]):
    """Implements endpoint for applying for a new certificate, that is ``/acme/new-order``.

    This is the first request of an ACME client when requesting a new certificate.

    If the client receives a successful response, it will next fetch the authorizations listed in it, which
    are served by :py:class:`~django_ca.views.AcmeAuthorizationView`.

    ``certbot`` sends the :py:class:`~acme:acme.messages.NewOrder` message via
    :py:meth:`~acme:acme.client.ClientV2.new_order`.

    .. seealso:: `RFC 8555, 7.4 <https://tools.ietf.org/html/rfc8555#section-7.4>`_
    """

    message_cls = NewOrder

    @transaction.atomic
    def acme_request(self, message: NewOrder, slug: Optional[str] = None) -> AcmeResponseOrderCreated:
        """Process ACME request."""
        now = datetime.now(tz.utc)

        # josepy message classes define field names as class variables, but instance attributes are of the
        # same type (similar to Django). So we cast josepy fields to the actual type.
        not_before = cast(Optional[datetime], message.not_before)
        not_after = cast(Optional[datetime], message.not_after)
        # TODO: test if identifiers are acceptable
        identifiers = typing.cast(List[messages.Identifier], message.identifiers)

        if not_before and not_before < now:
            raise AcmeMalformed(message="Certificate cannot be valid before now.")
        if not_after and not_after > now + ca_settings.ACME_MAX_CERT_VALIDITY:
            raise AcmeMalformed(message="Certificate cannot be valid that long.")
        if not_before and not_after and not_before > not_after:
            raise AcmeMalformed(message="notBefore must be before notAfter.")
        if not identifiers:
            # NOTE: Catches sending an empty tuple, which is not caught in message deserialization
            raise AcmeMalformed(message="The following fields are required: identifiers")

        if settings.USE_TZ is False:
            if not_before is not None and timezone.is_aware(not_before):
                not_before = timezone.make_naive(not_before)
            if not_after is not None and timezone.is_aware(not_after):
                not_after = timezone.make_naive(not_after)

        order = AcmeOrder.objects.create(account=self.account, not_before=not_before, not_after=not_after)
        authorizations = [
            self.request.build_absolute_uri(authz.acme_url) for authz in order.add_authorizations(identifiers)
        ]

        expires = order.expires
        if expires.tzinfo is None:  # acme.messages.Order requires a timezone-aware object
            expires = expires.replace(tzinfo=tz.utc)

        response = AcmeResponseOrderCreated(
            authorizations=authorizations,
            expires=expires,
            finalize=self.request.build_absolute_uri(order.acme_finalize_url),
            identifiers=identifiers,
            not_after=message.not_after,
            not_before=message.not_before,
            status=order.status,
        )
        response["Location"] = self.request.build_absolute_uri(order.acme_url)
        return response


class AcmeOrderView(AcmePostAsGetView):
    """Implements endpoint for viewing an order, that is ``/acme/order/<slug>/``.

    A client calls this view after calling :py:class:`~django_ca.views.AcmeOrderFinalizeView`, presumably
    to retrieve final information about this certificate. This view itself is sparsely documented in RFC 8555.

    This view does seem to be a little redundant too, as the response is almost identical to the response the
    client received in the previous request. Nevertheless, certbot still calls this view and fails without it.

    .. seealso:: `RFC 8555, 7.4 <https://tools.ietf.org/html/rfc8555#section-7.4>`_
    """

    def acme_request(self, slug: str) -> AcmeResponseOrder:
        try:
            order = AcmeOrder.objects.viewable().account(self.account).get(slug=slug)
        except AcmeOrder.DoesNotExist as ex:
            # RFC 8555, section 10.5: Avoid leaking info that this slug does not exist by
            # return a normal unauthorized message.
            raise AcmeUnauthorized() from ex
        # self.prepared['order'] = order.slug

        expires = order.expires
        if expires.tzinfo is None:  # acme.messages.Order requires a timezone-aware object
            expires = expires.replace(tzinfo=tz.utc)

        authorizations = order.authorizations.all()
        if order.status in [AcmeOrder.STATUS_VALID, AcmeOrder.STATUS_INVALID]:
            # RFC 8555, section 7.1.3:
            #
            #   For final orders (in the "valid" or "invalid" state), the authorizations that were completed.
            authorizations = authorizations.filter(status=AcmeAuthorization.STATUS_VALID)

        cert_url = None
        try:
            cert = AcmeCertificate.objects.get(order=order)
            if cert.cert and order.status == AcmeOrder.STATUS_VALID:
                # WARNING: certbot (at least version 0.31.0) will try to fetch the certificate immediately if
                # we return the URL. That view will fail if the certificate is not yet issued, and certbot
                # fails with an error.
                # This behavior is independent of the status of the order, despite the fact this is the field
                # that should be used for this according to RFC 8555.
                cert_url = self.request.build_absolute_uri(cert.acme_url)
        except AcmeCertificate.DoesNotExist:
            pass

        response = AcmeResponseOrder(
            status=order.status,
            expires=expires,
            identifiers=tuple({"type": a.type, "value": a.value} for a in authorizations),
            authorizations=tuple(self.request.build_absolute_uri(a.acme_url) for a in authorizations),
            certificate=cert_url,
        )
        response["Location"] = self.request.build_absolute_uri(order.acme_url)
        return response


class AcmeOrderFinalizeView(AcmeMessageBaseView[CertificateRequest]):
    """Implements endpoint for applying for certificate issuance, that is ``/acme/order/<slug>/finalize``.

    The client is supposed to call this URL to submit its CSR, once "it believes it has fulfilled the server's
    requirements".

    Note that in practice, the client can call this endpoint only once while the order is "ready". The
    endpoint returns an error if the order is not ready, and the call updates the state to "processing".

    .. seealso:: `RFC 8555, 7.4 <https://tools.ietf.org/html/rfc8555#section-7.4>`_
    """

    message_cls = CertificateRequest

    def validate_csr(self, message: CertificateRequest, authorizations: Iterable[AcmeAuthorization]) -> str:
        """Parse and validate the CSR, returns the PEM as str."""

        # Note: Jose wraps the CSR in a josepy.util.ComparableX509, that has *no* public member methods.
        # The only public attribute or function is the wrapped object. We encode it back to get the regular
        # PEM.
        # Note that the CSR received here is not an actual PEM, see AcmeCertificate.parse_csr()
        try:
            csr = parse_acme_csr(message.encode("csr"))
        except x509.InvalidVersion as ex:
            raise AcmeBadCSR(message="Invalid CSR version.") from ex
        except Exception as ex:
            log.exception("Error parsing CSR.")
            raise AcmeBadCSR(message="Unable to parse CSR.") from ex

        # Do not accept MD5 or SHA1 signatures
        hash_algorithm = csr.signature_hash_algorithm
        if hasattr(hashes, "MD5") and isinstance(hash_algorithm, hashes.MD5):
            raise AcmeBadCSR(message=f"{hash_algorithm.name}: Insecure hash algorithm.")
        if hasattr(hashes, "SHA1") and isinstance(hash_algorithm, hashes.SHA1):
            raise AcmeBadCSR(message=f"{hash_algorithm.name}: Insecure hash algorithm.")

        if csr.is_signature_valid is False:
            raise AcmeBadCSR(message="CSR signature is not valid.")

        # Get list of general names from the authorizations
        names_from_order = set(auth.general_name for auth in authorizations)

        # Perform sanity checks on the CSRs subject.
        # NOTE: certbot does *not* set any subject at all
        if csr.subject:
            check_name(csr.subject)

            # We allow a client setting a CommonName, but it *must* be part of the order.
            common_name = next((attr for attr in csr.subject if attr.oid == NameOID.COMMON_NAME), None)
            if common_name is not None:
                if isinstance(common_name.value, bytes):  # pragma: no cover
                    raise AcmeBadCSR(message="CommonName was not in order.")
                if x509.DNSName(common_name.value) not in names_from_order:
                    raise AcmeBadCSR(message="CommonName was not in order.")

        try:
            san_ext = csr.extensions.get_extension_for_class(x509.SubjectAlternativeName)
            names_from_csr: Set[x509.GeneralName] = set(san_ext.value)
        except x509.ExtensionNotFound as ex:
            raise AcmeBadCSR(message="No subject alternative names found in CSR.") from ex

        if names_from_order != names_from_csr:
            raise AcmeBadCSR(message="Names in CSR do not match.")

        return csr.public_bytes(Encoding.PEM).decode("utf-8")

    def acme_request(self, message: CertificateRequest, slug: Optional[str]) -> AcmeResponseOrder:
        """Process ACME request."""
        try:
            order = AcmeOrder.objects.viewable().account(account=self.account).get(slug=slug)
        except AcmeOrder.DoesNotExist as ex:
            # RFC 8555, section 10.5: Avoid leaking info that this slug does not exist by
            # return a normal unauthorized message.
            raise AcmeUnauthorized() from ex
        # self.prepared['order'] = order.slug

        # RFC 8555, section 7.4:
        #
        #   A request to finalize an order will result in error if the order is not in the "ready" state.  In
        #   such cases, the server MUST return a 403 (Forbidden) error with a problem document of type
        #   "orderNotReady".  The client should then send a POST-as-GET request to the order resource to
        #   obtain its current state.  The status of the order will indicate what action the client should
        #   take (see below).
        if order.status != AcmeOrder.STATUS_READY:
            # NOTE: The provision quoted above means that you will *always* get "orderNotReady", even if you
            # fetch this URL and the certificate has already been issued. We might consider returning the
            # order instead in this case.
            # The spec also says you should send a POST-as-GET after a certain time, if the order is in the
            # processing state, but it's not entirely clear if that request should go here or the normal order
            # resource.
            # Further investigation is on what LE and certbot do is needed here.
            raise AcmeForbidden(typ="orderNotReady", message="This order is not yet ready.")

        expires = order.expires
        if expires.tzinfo is None:  # acme.messages.Order requires a timezone-aware object
            expires = expires.replace(tzinfo=tz.utc)

        authorizations = order.authorizations.all()
        for auth in authorizations:
            if auth.status != AcmeAuthorization.STATUS_VALID:
                # This is a state that should never happen in practice, because the order is only marked as
                # ready once all authorizations are valid.
                raise AcmeForbidden(typ="orderNotReady", message="This order is not yet ready.")

        # TODO: Do we validate that the account holds ALL necessary authorizations?

        # Parse and validate the CSR
        csr = self.validate_csr(message, authorizations)

        # Create AcmeCertificate object (at this point without cert, as it hasn't been issued yet)
        cert = AcmeCertificate.objects.create(order=order, csr=csr)

        # Update the status of the order to "processing"
        order.status = AcmeOrder.STATUS_PROCESSING
        order.save()

        # start task only after commit, see:
        # https://docs.djangoproject.com/en/dev/topics/db/transactions/#django.db.transaction.on_commit
        transaction.on_commit(lambda: run_task(acme_issue_certificate, acme_certificate_pk=cert.pk))

        response = AcmeResponseOrder(
            status=order.status,
            expires=expires,
            identifiers=tuple({"type": a.type, "value": a.value} for a in authorizations),
            authorizations=tuple(self.request.build_absolute_uri(a.acme_url) for a in authorizations),
        )
        response["Location"] = self.request.build_absolute_uri(order.acme_url)
        return response


class AcmeCertificateView(AcmePostAsGetView):
    """Implements endpoint to download a certificate, that is ``/acme/cert/<slug>/``.

    This is the final view called in the certificate validation process and downloads the issued certificate.

    .. seealso:: `RFC8555, 8555, 7.4.2 <https://tools.ietf.org/html/rfc8555#section-7.4.2>`_
    """

    # This is the only view that does not return JSON, thus acme_request() returns the superclass
    # HttpResponse, and not an AcmeResponse (which is always JSON).
    def acme_request(self, slug: str) -> HttpResponse:  # type: ignore[override]
        try:
            cert = AcmeCertificate.objects.viewable().account(self.account).get(slug=slug)
        except AcmeCertificate.DoesNotExist as ex:
            raise AcmeUnauthorized() from ex

        # COVERAGE NOTE: This should not happen in practice, as viewable() above already excludes those
        # instances. We add this check here just to be sure (and to make mypy happy).
        if cert.cert is None:  # pragma: no cover
            raise AcmeUnauthorized()

        # self.prepared['cert'] = slug
        # self.prepared['csr'] = cert.csr
        # self.prepared['order'] = cert.order.slug
        return HttpResponse(cert.cert.bundle_as_pem, content_type="application/pem-certificate-chain")


class AcmeAuthorizationView(AcmePostAsGetView):
    """Implements endpoint for identifier authorization, that is ``/acme/authz/<slug>/``.

    This is the second request when a client requests a new certificate and represents an authorization
    request for one of the identifiers in the order. The URL for this view is returned by the
    :py:class:`~django_ca.views.AcmeNewOrderView`.

    This view returns URLs to the challenges served by ::py:class:`~django_ca.views.AcmeChallengeView`. The
    client will then post to one of these challenges to start the validation process.

    The client periodically polls this view after initiating a challenge to know when the server has
    successfully validated the challenge.

    Note that this resource accepts both a post-as-get request but also a response body: RFC 8555, section
    7.5.2 describes deactivating an account authorization.

    .. seealso:: `RFC 8555, 7.5 <https://tools.ietf.org/html/rfc8555#section-7.5>`_
    """

    def acme_request(self, slug: str) -> AcmeResponseAuthorization:
        # TODO: implement deactivating an authorization (section 7.5.2)

        try:
            auth = AcmeAuthorization.objects.viewable().account(account=self.account).url().get(slug=slug)
        except AcmeAuthorization.DoesNotExist as ex:
            # RFC 8555, section 10.5: Avoid leaking info that this slug does not exist by
            # return a normal unauthorized message.
            raise AcmeUnauthorized() from ex

        # self.prepared['order'] = auth.order.slug
        # self.prepared['auth'] = auth.slug
        challenges = auth.get_challenges()

        expires = auth.expires
        if expires.tzinfo is None:  # acme.Order requires a timezone-aware object
            expires = expires.replace(tzinfo=tz.utc)

        # RFC8555, section 7.5.1:
        #
        #   "When finalizing an authorization, the server MAY remove challenges other than the one that was
        #   completed".
        #
        # The example response at the end of section 7.5.1 also only shows the valid challenge.
        if auth.status == AcmeAuthorization.STATUS_VALID:
            challenges = [c for c in challenges if c.status == AcmeChallenge.STATUS_VALID]

        resp = AcmeResponseAuthorization(
            identifier=auth.identifier,
            challenges=tuple(c.get_challenge(self.request) for c in challenges),
            status=auth.status,
            expires=expires,
        )
        return resp


class AcmeChallengeView(AcmePostAsGetView):
    """Implements ``/acme/chall/<slug>``, indicating to the server that the challenge can now be validated.

    The client calls this view to tell the server to start validation of the resource of this challenges
    resource using the challenge method (http-01, dns-01, ...) for this challenge.

    After this view is called, the client will poll :py:class:`~django_ca.views.AcmeAuthorizationView` to know
    when the server has validated the challenge. After successful validation the client calls
    :py:class:`~django_ca.views.AcmeOrderFinalizeView` to request issuing of the certificate.

    .. seealso::

        * `RFC 8555, section 7.1.5 <https://tools.ietf.org/html/rfc8555#section-7.1.5>`_
        * `RFC 8555, section 7.5.1 <https://tools.ietf.org/html/rfc8555#section-7.5.1>`_
    """

    ignore_body = True

    def set_link_relations(self, response: "HttpResponseBase", **kwargs: str) -> None:
        """Set the "up" link header to the matching authorization.

        `RFC8555, section 7.1 <https://tools.ietf.org/html/rfc8555#section-7.1>`_ states:

            The "up" link relation is used with challenge resources to indicate the authorization resource to
            which a challenge belongs.
        """
        if response.status_code < HTTPStatus.BAD_REQUEST:
            # Only return an up relation if no error is thrown
            kwargs["up"] = self.auth.acme_url
        super().set_link_relations(response, **kwargs)

    def acme_request(self, slug: str) -> AcmeResponseChallenge:
        try:
            challenge = AcmeChallenge.objects.viewable().account(self.account).url().get(slug=slug)
        except AcmeChallenge.DoesNotExist:
            # RFC 8555, section 10.5: Avoid leaking info that this slug does not exist by
            # return a normal unauthorized message.
            raise AcmeUnauthorized()  # pylint: disable=raise-missing-from

        # Set self.auth attribute, we need it in set_link_relations()
        self.auth = challenge.auth  # pylint: disable=attribute-defined-outside-init

        # self.prepared['order'] = challenge.auth.order.slug
        # self.prepared['auth'] = challenge.auth.slug
        # self.prepared['challenge'] = slug

        if challenge.usable is True:  # if not -> no state change
            # RFC8555, Section 7.1.6:
            #
            #   They transition to the "processing" state when the client responds to the challenge
            challenge.status = AcmeChallenge.STATUS_PROCESSING
            challenge.save()

            # Actually perform challenge validation asynchronously
            # start task only after commit, see:
            # https://docs.djangoproject.com/en/2.2/topics/db/transactions/#django.db.transaction.on_commit
            transaction.on_commit(lambda: run_task(acme_validate_challenge, challenge.pk))

        return AcmeResponseChallenge(
            chall=challenge.acme_challenge,
            _url=self.request.build_absolute_uri(challenge.acme_url),
            status=challenge.status,
            validated=challenge.validated,
        )


class AcmeCertificateRevocationView(AcmeMessageBaseView[messages.Revocation]):
    """View providing ACMEv2 certificate revocation.

    .. seealso::

        `RFC 8555, section 7.6 <https://tools.ietf.org/html/rfc8555#section-7.6>`_
    """

    accepts_kid_or_jwk = True
    message_cls = messages.Revocation

    def get_certificate(self, serial: str) -> Certificate:
        """Get the certificate that is to be revoked by this request.

        This function handles the special authorization requirements for this request (they can be signed by
        either the account key pair or the certificate key pair).
        """

        certs = Certificate.objects.filter(ca=self.ca).currently_valid()

        # If the request is signed with the certificate key (and not the account), a JWK is set for this
        # request, and we verify it was signed by the certificate on record.
        if self.jws.signature.combined.jwk:
            # NOTE: RFC 8555, section 7.6 states:
            #
            #   The server MUST also consider a revocation request valid if it is signed with the private key
            #   corresponding to the public key in the certificate.
            #
            # This implies that the account used to request the certificate may even be revoked or invalid, as
            # long as the private key is used to sign the request. So we don't look at the ACME account at all
            # here.
            cert: Certificate = certs.get(serial=serial)

            jwk = cert.jwk

            # The JWS signature was already verified using the JWK from the request in the base class. But we
            # still need to verify that it was signed by the certificate that is to be revoked (and not just
            # any certificate).
            #
            # Theoretically, cert.jwk != self.jwk would be a sufficient check, but we do not trust the
            # __eq__ implementation given the linked GitHub issues, so we also verify the signature a second
            # time based on the database key.
            #   https://josepy.readthedocs.io/en/latest/api/util/
            if jwk != self.jwk or not self.jws.verify(jwk):
                raise AcmeUnauthorized(message="Request signed by the wrong certificate.")
        else:
            # Get the certificate by serial if it *has* an ACME account.
            # NOTE: The base class already makes sure that the account is currently valid.
            cert = certs.filter(acmecertificate__order__account__isnull=False).get(serial=serial)

            # If the request is from the account that issued the certificate, the certificate can be revoked.
            # NOTE: self.account is **only** set if the request has no JWK.
            if cert.acmecertificate.order.account == self.account:
                return cert

            # If the account holds authorizations for all the identifiers in the certificate, it can also
            # be revoked, so get a list of all currently valid authorizations that the account holds
            authz = set(AcmeAuthorization.objects.dns().valid().account(account=self.account).names())

            # Get names from the certificate, first from the CommonName...
            # NOTE: returns empty list if subject does not have a CommonName.
            names = [cn.value for cn in cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)]

            # ... then from the SubjectAlternativeName extension
            try:
                san = cert.pub.loaded.extensions.get_extension_for_class(x509.SubjectAlternativeName)

                # If ANY subjectAlternativeName is NOT a DNS name, we cannot revoke this cert.
                if list(filter(lambda v: not isinstance(v, x509.DNSName), san.value)):
                    raise AcmeUnauthorized(message="Certificate contains non-DNS subjectAlternativeNames.")
                names += [name.value for name in san.value]
            except x509.ExtensionNotFound:
                pass

            # Finally test if the account holds all authorizations required for revoking this certificate.
            if not set(names) <= authz:
                raise AcmeUnauthorized(message="Account does not hold necessary authorizations.")

        return cert

    def acme_request(self, message: messages.Revocation, slug: Optional[str]) -> AcmeResponse:
        reason_code = message.reason
        if reason_code is None:
            reason_code = 0

        try:
            reason = REASON_CODES[reason_code]
        except KeyError as ex:
            raise AcmeMalformed(
                typ="badRevocationReason", message=f"{message.reason}: Unsupported revocation reason."
            ) from ex

        # Get cryptography certificate from ACME message
        cg_cert = message.certificate.wrapped.to_cryptography()
        if not isinstance(cg_cert, x509.Certificate):  # pragma: no cover
            # COVERAGE NOTE: message deserialization already raises an error when no certificate is passed,
            # so this check here is just for more safety (and to make mypy happy).
            raise AcmeMalformed(message="Request did not contain a certificate.")

        try:
            cert = self.get_certificate(int_to_hex(cg_cert.serial_number))
        except Certificate.DoesNotExist as ex:
            raise AcmeUnauthorized(message="Certificate not found.") from ex

        # Check that the certificate in the payload matches with the one on record
        if cert.pub.loaded != cg_cert:
            raise AcmeUnauthorized(message="Certificate does not match records.")

        # RFC 8555, section 7.6
        #
        #   if the certificate has already been revoked, the server returns an error response with status code
        #   400 (Bad Request) and type "urn:ietf:params:acme:error:alreadyRevoked".
        if cert.revoked:
            raise AcmeMalformed(typ="alreadyRevoked", message="Certificate was already revoked.")

        # Finally actually revoke the certificate
        cert.revoke(reason)
        return AcmeResponse({})  # No response specified in RFC 8555!



# Generated by Django 3.2 on 2021-04-30 11:31

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0023_auto_20210429_0000'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='certificate',
            name='csr',
        ),
        migrations.RemoveField(
            model_name='certificate',
            name='pub',
        ),
        migrations.RemoveField(
            model_name='certificateauthority',
            name='pub',
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to view details for a certificate.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from typing import Any

from django_ca.management.base import BaseViewCommand
from django_ca.management.mixins import CertCommandMixin
from django_ca.models import Certificate


class Command(CertCommandMixin, BaseViewCommand):
    """Implement :command:`manage.py view_cert`."""

    allow_revoked = True
    help = 'View a certificate. The "list_certs" command lists all known certificates.'

    def handle(self, cert: Certificate, pem: bool, extensions: bool, wrap: bool, **options: Any) -> None:
        self.output_header(cert)

        watchers = cert.watchers.all()
        if watchers:
            self.stdout.write("* Watchers:")
            for watcher in watchers:
                self.stdout.write(f"  * {watcher}")
        else:
            self.stdout.write("* No watchers")

        # self.stdout.write extensions
        if extensions:
            self.stdout.write("\nCertificate extensions:")
            self.print_extensions(cert)

        self.output_footer(cert, pem=pem, wrap=wrap)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""``django_ca.extensions.utils`` contains various utility classes used by X.509 extensions."""

import typing
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from cryptography import x509
from cryptography.x509.certificate_transparency import LogEntryType, SignedCertificateTimestamp
from cryptography.x509.oid import ExtensionOID

from django.template.loader import render_to_string
from django.urls import reverse

from django_ca.constants import KEY_USAGE_NAMES
from django_ca.typehints import ExtensionMapping
from django_ca.utils import add_colons, bytes_to_hex, int_to_hex


def extension_as_admin_html(
    extension: x509.Extension[x509.ExtensionType], extra_context: Optional[Dict[str, Any]] = None
) -> str:
    """Convert an extension to HTML code suitable for the admin interface."""

    template = f"django_ca/admin/extensions/{extension.oid.dotted_string}.html"
    if isinstance(extension.value, x509.UnrecognizedExtension):
        template = "django_ca/admin/extensions/unrecognized_extension.html"

    context = {"extension": extension, "x509": x509}
    if extra_context is not None:
        context.update(extra_context)

    return render_to_string(template, context=context)


def certificate_policies_is_simple(value: x509.CertificatePolicies) -> bool:
    """Check if a Certificate Policies extension is "simple".

    The extension is considered simple if it contains a single Policy Information, and that policy information
    contains no notice references and at most one explicit text.
    """
    if len(value) > 1:
        return False

    policy_information = value[0]
    has_explicit_text = False

    if not policy_information.policy_qualifiers:
        return True

    for policy_qualifier in policy_information.policy_qualifiers:
        if isinstance(policy_qualifier, str):
            continue  # we support multiple - no need to keep track

        if policy_qualifier.notice_reference:
            return False
        if has_explicit_text:
            return False
        has_explicit_text = True
    return True


def key_usage_items(value: x509.KeyUsage) -> Iterator[str]:
    """Get a list of basic key usages."""
    for attr, name in KEY_USAGE_NAMES.items():
        try:
            if getattr(value, attr):
                yield name
        except ValueError:
            # x509.KeyUsage raises ValueError on some attributes to ensure consistency
            pass


def signed_certificate_timestamp_values(sct: SignedCertificateTimestamp) -> Tuple[str, str, str, str]:
    """Get values from a SignedCertificateTimestamp as a tuple of strings."""
    if sct.entry_type == LogEntryType.PRE_CERTIFICATE:
        entry_type = "Precertificate"
    elif sct.entry_type == LogEntryType.X509_CERTIFICATE:  # pragma: no cover  # Unseen in the wild
        entry_type = "x509 certificate"
    else:  # pragma: no cover  # We support everything that has been specified so far
        entry_type = "unknown"
    return entry_type, sct.version.name, bytes_to_hex(sct.log_id), sct.timestamp.isoformat(" ")


def get_formatting_context(serial: int, signer_serial: int) -> Dict[str, Union[int, str]]:
    """Get the context for formatting extensions."""
    hex_serial = int_to_hex(serial)
    signer_serial_hex = int_to_hex(signer_serial)
    return {
        "SERIAL": serial,
        "SERIAL_HEX": hex_serial,
        "SERIAL_HEX_COLONS": add_colons(hex_serial),
        "SIGNER_SERIAL": signer_serial,
        "SIGNER_SERIAL_HEX": signer_serial_hex,
        "SIGNER_SERIAL_HEX_COLONS": add_colons(signer_serial_hex),
        "CA_ISSUER_PATH": reverse("django_ca:issuer", kwargs={"serial": signer_serial_hex}).lstrip("/"),
    }


def format_general_name(name: x509.GeneralName, context: Dict[str, Union[str, int]]) -> x509.GeneralName:
    """Format a general name (currently only operating on UniformResourceIdentifier)."""
    if isinstance(name, x509.UniformResourceIdentifier):
        return x509.UniformResourceIdentifier(name.value.format(**context))
    return name


def format_extensions(extensions: ExtensionMapping, context: Dict[str, Union[str, int]]) -> None:
    """Format extensions based on the given context."""
    if ExtensionOID.AUTHORITY_INFORMATION_ACCESS in extensions:
        authority_information_access = typing.cast(
            x509.Extension[x509.AuthorityInformationAccess],
            extensions[ExtensionOID.AUTHORITY_INFORMATION_ACCESS],
        )

        access_descriptions = [
            x509.AccessDescription(
                access_method=ad.access_method,
                access_location=format_general_name(ad.access_location, context),
            )
            for ad in authority_information_access.value
        ]
        extensions[ExtensionOID.AUTHORITY_INFORMATION_ACCESS] = x509.Extension(
            oid=ExtensionOID.AUTHORITY_INFORMATION_ACCESS,
            critical=authority_information_access.critical,
            value=x509.AuthorityInformationAccess(access_descriptions),
        )

    if ExtensionOID.CRL_DISTRIBUTION_POINTS in extensions:
        crl_distribution_points = typing.cast(
            x509.Extension[x509.CRLDistributionPoints],
            extensions[ExtensionOID.CRL_DISTRIBUTION_POINTS],
        )

        distribution_points: List[x509.DistributionPoint] = []

        distribution_point: x509.DistributionPoint
        for distribution_point in crl_distribution_points.value:
            if distribution_point.full_name is None:
                distribution_points.append(distribution_point)
            else:
                names = [format_general_name(name, context) for name in distribution_point.full_name]
                distribution_points.append(
                    x509.DistributionPoint(
                        full_name=names,
                        relative_name=None,
                        reasons=distribution_point.reasons,
                        crl_issuer=distribution_point.crl_issuer,
                    )
                )

        extensions[ExtensionOID.CRL_DISTRIBUTION_POINTS] = x509.Extension(
            oid=ExtensionOID.CRL_DISTRIBUTION_POINTS,
            critical=crl_distribution_points.critical,
            value=x509.CRLDistributionPoints(distribution_points),
        )



# Generated by Django 3.0.2 on 2020-01-12 16:57

from django.db import migrations, models
import django_ca.utils


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0016_auto_20190706_1548'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certificate',
            name='revoked_reason',
            field=models.CharField(blank=True, choices=[('aa_compromise', 'Attribute Authority compromised'), ('affiliation_changed', 'Affiliation changed'), ('ca_compromise', 'CA compromised'), ('certificate_hold', 'On Hold'), ('cessation_of_operation', 'Cessation of operation'), ('key_compromise', 'Key compromised'), ('privilege_withdrawn', 'Privilege withdrawn'), ('remove_from_crl', 'Removed from CRL'), ('superseded', 'Superseded'), ('unspecified', 'Unspecified')], default='', max_length=32, verbose_name='Reason for revokation'),
        ),
        migrations.AlterField(
            model_name='certificateauthority',
            name='crl_url',
            field=models.TextField(blank=True, default='', help_text='URLs, one per line, where you can retrieve the CRL.', validators=[django_ca.utils.multiline_url_validator], verbose_name='CRL URLs'),
        ),
        migrations.AlterField(
            model_name='certificateauthority',
            name='revoked_reason',
            field=models.CharField(blank=True, choices=[('aa_compromise', 'Attribute Authority compromised'), ('affiliation_changed', 'Affiliation changed'), ('ca_compromise', 'CA compromised'), ('certificate_hold', 'On Hold'), ('cessation_of_operation', 'Cessation of operation'), ('key_compromise', 'Key compromised'), ('privilege_withdrawn', 'Privilege withdrawn'), ('remove_from_crl', 'Removed from CRL'), ('superseded', 'Superseded'), ('unspecified', 'Unspecified')], default='', max_length=32, verbose_name='Reason for revokation'),
        ),
        migrations.AlterField(
            model_name='watcher',
            name='name',
            field=models.CharField(blank=True, default='', max_length=64, verbose_name='CommonName'),
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Collection of exception classes for ACMEv2."""

from typing import Any, Type

from django_ca.acme.responses import (
    AcmeResponseBadCSR,
    AcmeResponseError,
    AcmeResponseForbidden,
    AcmeResponseMalformed,
    AcmeResponseUnauthorized,
)


class AcmeException(Exception):
    """Base class for all ACME exceptions."""

    response: Type[AcmeResponseError] = AcmeResponseError

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args)
        self.kwargs = kwargs

    def get_response(self) -> AcmeResponseError:
        """Get the corresponding ACME response class."""
        return self.response(*self.args, **self.kwargs)


class AcmeMalformed(AcmeException):
    """Exception when the request was malformed."""

    response: Type[AcmeResponseMalformed] = AcmeResponseMalformed


class AcmeUnauthorized(AcmeException):
    """Exception when the request is unauthorized."""

    response = AcmeResponseUnauthorized  # 401


class AcmeForbidden(AcmeException):
    """Exception when the request is unauthorized."""

    response = AcmeResponseForbidden  # 403


class AcmeBadCSR(AcmeException):
    """Exception raised when a CSR is not acceptable."""

    response = AcmeResponseBadCSR



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Django model managers."""


import pathlib
import typing
from typing import Any, Dict, Generic, Iterable, List, Optional, TypeVar, Union

from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat
from cryptography.x509.oid import AuthorityInformationAccessOID, ExtensionOID

from django.core.files.base import ContentFile
from django.db import models
from django.urls import reverse

from django_ca import ca_settings, constants
from django_ca.extensions.utils import format_extensions, get_formatting_context
from django_ca.modelfields import LazyCertificateSigningRequest
from django_ca.openssh import SshHostCaExtension, SshUserCaExtension
from django_ca.profiles import Profile, profiles
from django_ca.signals import post_create_ca, post_issue_cert, pre_create_ca
from django_ca.typehints import (
    AllowedHashTypes,
    Expires,
    ExtensionMapping,
    ParsableKeyType,
    X509CertMixinTypeVar,
)
from django_ca.utils import (
    ca_storage,
    format_general_name,
    generate_private_key,
    get_cert_builder,
    parse_expires,
    validate_hostname,
    validate_private_key_parameters,
    validate_public_key_parameters,
)

# https://mypy.readthedocs.io/en/latest/runtime_troubles.html
if typing.TYPE_CHECKING:
    from django_ca.models import (
        AcmeAccount,
        AcmeAuthorization,
        AcmeCertificate,
        AcmeChallenge,
        AcmeOrder,
        Certificate,
        CertificateAuthority,
    )
    from django_ca.querysets import AcmeAccountQuerySet, CertificateAuthorityQuerySet, CertificateQuerySet

    AcmeAccountManagerBase = models.Manager[AcmeAccount]
    AcmeAuthorizationManagerBase = models.Manager[AcmeAuthorization]
    AcmeCertificateManagerBase = models.Manager[AcmeCertificate]
    AcmeChallengeManagerBase = models.Manager[AcmeChallenge]
    AcmeOrderManagerBase = models.Manager[AcmeOrder]
    CertificateAuthorityManagerBase = models.Manager[CertificateAuthority]
    CertificateManagerBase = models.Manager[Certificate]

    QuerySetTypeVar = TypeVar("QuerySetTypeVar", CertificateAuthorityQuerySet, CertificateQuerySet)
else:
    AcmeAccountManagerBase = (
        AcmeAuthorizationManagerBase
    ) = (
        AcmeCertificateManagerBase
    ) = (
        AcmeChallengeManagerBase
    ) = AcmeOrderManagerBase = CertificateAuthorityManagerBase = CertificateManagerBase = models.Manager
    QuerySetTypeVar = TypeVar("QuerySetTypeVar")


class CertificateManagerMixin(Generic[X509CertMixinTypeVar, QuerySetTypeVar]):
    """Mixin for model managers."""

    if typing.TYPE_CHECKING:
        # django-stubs (mypy plugin for Django) currently typehints queryset methods as returning a manager,
        # and does not know about queryset methods coming from the queryset. We typehint basic queryset
        # methods here, so that mypy knows that returned objects are querysets.
        #
        # The type overrides are because of the return type, as mypy thinks they should return a manager.
        #
        # pylint: disable=missing-function-docstring,unused-argument; just defining stubs here

        def all(self) -> QuerySetTypeVar:
            ...

        def get_queryset(self) -> QuerySetTypeVar:
            ...

        def filter(self, *args: Any, **kwargs: Any) -> QuerySetTypeVar:
            ...

        def exclude(self, *args: Any, **kwargs: Any) -> QuerySetTypeVar:
            ...

        def order_by(self, *fields: str) -> QuerySetTypeVar:
            ...

        def get_by_serial_or_cn(self, identifier: str) -> X509CertMixinTypeVar:
            ...

        def valid(self) -> QuerySetTypeVar:
            ...


class CertificateAuthorityManager(
    CertificateManagerMixin["CertificateAuthority", "CertificateAuthorityQuerySet"],
    CertificateAuthorityManagerBase,
):
    """Model manager for the CertificateAuthority model."""

    if typing.TYPE_CHECKING:
        # See CertificateManagerMixin for description on this branch
        #
        # pylint: disable=missing-function-docstring; just defining stubs here

        def acme(self) -> "CertificateAuthorityQuerySet":
            ...

        def default(self) -> "CertificateAuthority":
            ...

        def disabled(self) -> "CertificateAuthorityQuerySet":
            ...

        def enabled(self) -> "CertificateAuthorityQuerySet":
            ...

        def invalid(self) -> "CertificateAuthorityQuerySet":
            ...

        def usable(self) -> "CertificateAuthorityQuerySet":
            ...

    def _get_formatting_context(self, serial: int, signer_serial: int) -> Dict[str, Union[int, str]]:
        context = get_formatting_context(serial, signer_serial)
        kwargs = {"serial": context["SIGNER_SERIAL_HEX"]}
        context["OCSP_PATH"] = reverse("django_ca:ocsp-ca-post", kwargs=kwargs).lstrip("/")
        context["CRL_PATH"] = reverse("django_ca:ca-crl", kwargs=kwargs).lstrip("/")
        return context

    def _handle_authority_information_access(
        self,
        hostname: Optional[str],
        extensions: ExtensionMapping,
    ) -> None:
        """Add an Authority Information Access extension with a URI based on `hostname` to `extensions`.

        If the extension is already present, OCSP/CA Issuers access description are only added if the
        extension does not contain any access descriptions of the respective type.
        """
        oid = ExtensionOID.AUTHORITY_INFORMATION_ACCESS
        access_descriptions: List[x509.AccessDescription] = []
        if oid in extensions:
            extension = typing.cast(x509.AuthorityInformationAccess, extensions[oid].value)
            access_descriptions = list(extension)

        has_ocsp = any(ad.access_method == AuthorityInformationAccessOID.OCSP for ad in access_descriptions)
        has_issuer = any(
            ad.access_method == AuthorityInformationAccessOID.CA_ISSUERS for ad in access_descriptions
        )

        # Fields are only added if not already present, so if both are present, we have nothing to do
        if has_ocsp and has_issuer:
            return

        if has_issuer is False:
            access_descriptions.append(
                x509.AccessDescription(
                    access_method=AuthorityInformationAccessOID.CA_ISSUERS,
                    access_location=x509.UniformResourceIdentifier(f"http://{hostname}/{{CA_ISSUER_PATH}}"),
                )
            )
        if has_ocsp is False:
            access_descriptions.append(
                x509.AccessDescription(
                    access_method=AuthorityInformationAccessOID.OCSP,
                    access_location=x509.UniformResourceIdentifier(f"http://{hostname}/{{OCSP_PATH}}"),
                )
            )

        # Finally sort by OID so that we have more predictable behavior
        access_descriptions = sorted(access_descriptions, key=lambda ad: ad.access_method.dotted_string)

        extensions[oid] = x509.Extension(
            oid=oid,
            critical=constants.EXTENSION_DEFAULT_CRITICAL[oid],
            value=x509.AuthorityInformationAccess(access_descriptions),
        )

    def _handle_crl_distribution_point(self, hostname: Optional[str], extensions: ExtensionMapping) -> None:
        """Add CRL Distribution Point extension with a URI based on `hostname` to `extensions`.

        The extension is only added if it is not already set in `extensions`.
        """
        oid = ExtensionOID.CRL_DISTRIBUTION_POINTS
        if oid in extensions:
            return

        uri = x509.UniformResourceIdentifier(f"http://{hostname}/{{CRL_PATH}}")
        extensions[oid] = x509.Extension(
            oid=oid,
            critical=constants.EXTENSION_DEFAULT_CRITICAL[oid],
            value=x509.CRLDistributionPoints(
                [x509.DistributionPoint(full_name=[uri], relative_name=None, reasons=None, crl_issuer=None)]
            ),
        )

    def init(
        self,
        name: str,
        subject: x509.Name,
        expires: Expires = None,
        algorithm: Optional[AllowedHashTypes] = None,
        parent: Optional["CertificateAuthority"] = None,
        default_hostname: Optional[Union[bool, str]] = None,
        path_length: Optional[int] = None,
        issuer_url: str = "",
        issuer_alt_name: Optional[x509.Extension[x509.IssuerAlternativeName]] = None,
        crl_url: Optional[Iterable[str]] = None,
        ocsp_url: str = "",
        password: Optional[Union[str, bytes]] = None,
        parent_password: Optional[Union[str, bytes]] = None,
        elliptic_curve: Optional[ec.EllipticCurve] = None,
        key_type: ParsableKeyType = "RSA",
        key_size: Optional[int] = None,
        extensions: Optional[Iterable[x509.Extension[x509.ExtensionType]]] = None,
        path: Union[pathlib.PurePath, str] = "ca",
        caa: str = "",
        website: str = "",
        terms_of_service: str = "",
        acme_enabled: bool = False,
        acme_registration: bool = True,
        acme_requires_contact: bool = True,
        acme_profile: Optional[str] = None,
        openssh_ca: bool = False,
        sign_certificate_policies: Optional[x509.Extension[x509.CertificatePolicies]] = None,
        ocsp_responder_key_validity: Optional[int] = None,
        ocsp_response_validity: Optional[int] = None,
    ) -> "CertificateAuthority":
        """Create a new certificate authority.

        .. versionchanged:: 1.26.0

           * The `permitted_subtrees` and `excluded_subtrees` subtrees where removed. Pass a
             :py:class:`~cg:cryptography.x509.NameConstraints` extension in `extensions` instead.
           * Added the `acme_registration` option.

        Parameters
        ----------

        name : str
            The name of the CA. This is a human-readable string and is used for administrative purposes only.
        subject : :py:class:`cg:cryptography.x509.Name`
           The desired subject for the certificate.
        expires : int or datetime or timedelta, optional
            When this certificate authority will expire, defaults to :ref:`CA_DEFAULT_EXPIRES
            <settings-ca-default-expires>`.
        algorithm : :py:class:`~cg:cryptography.hazmat.primitives.hashes.HashAlgorithm`, optional
            Hash algorithm used when signing the certificate, defaults to
            :ref:`settings-ca-default-signature-hash-algorithm` for RSA/EC keys, and
            :ref:`settings-ca-default-dsa-signature-hash-algorithm` for DSA keys. Passing an algorithm for
            Ed448/Ed25519 keys is an error.
        parent : :py:class:`~django_ca.models.CertificateAuthority`, optional
            Parent certificate authority for the new CA. Passing this value makes the CA an intermediate
            authority. Let unset if this CA will be used for OpenSSH.
        default_hostname : str, optional
            Override the URL configured with :ref:`CA_DEFAULT_HOSTNAME <settings-ca-default-hostname>` with a
            different hostname. Set to ``False`` to disable the hostname.
        path_length : int, optional
            Value of the path length attribute for the Basic Constraints extension.
        issuer_url : str
            URL for the DER/ASN1 formatted certificate that is signing certificates.
        issuer_alt_name : :py:class:`~cg:cryptography.x509.Extension`, optional
            IssuerAlternativeName used when signing certificates.  The value of the extension must be an
            :py:class:`~cg:cryptography.x509.IssuerAlternativeName` instance.
        crl_url : list of str, optional
            CRL URLs used for certificates signed by this CA.
        ocsp_url : str, optional
            OCSP URL used for certificates signed by this CA. The default is no value, unless
            :ref:`CA_DEFAULT_HOSTNAME <settings-ca-default-hostname>` is set.
        password : bytes or str, optional
            Password to encrypt the private key with.
        parent_password : bytes or str, optional
            Password that the private key of the parent CA is encrypted with.
        elliptic_curve : :py:class:`~cg:cryptography.hazmat.primitives.asymmetric.ec.EllipticCurve`, optional
            An elliptic curve to use for EC keys. This parameter is ignored if ``key_type`` is not ``"EC"``.
            Defaults to the :ref:`CA_DEFAULT_ELLIPTIC_CURVE <settings-ca-default-elliptic-curve>`.
        key_type: str, optional
            The type of private key to generate, must be one of ``"RSA"``, ``"DSA"``, ``"EC"``, or
            ``"Ed25519"`` , with ``"RSA"`` being the default.
        key_size : int, optional
            Integer specifying the key size, must be a power of two (e.g. 2048, 4096, ...). Defaults to
            the :ref:`CA_DEFAULT_KEY_SIZE <settings-ca-default-key-size>`, unused if
            ``key_type="EC"`` or ``key_type="Ed25519"``.
        extensions : list of :py:class:`cg:cryptography.x509.Extension`
            An optional list of additional extensions to add to the certificate.
        path : str or pathlib.PurePath, optional
            Where to store the CA private key (default ``ca``).
        caa : str, optional
            CAA identity. Note that this field is not yet currently used.
        website : str, optional
            URL to a human-readable website.
        terms_of_service : str, optional
            URL to the terms of service for the CA.
        acme_enabled : bool, optional
            Set to ``True`` to enable ACMEv2 support for this CA.
        acme_registration : bool, optional
            Whether to allow ACMEv2 clients to register new ACMEv2 accounts (if support is enabled in the
            first place). By default, account registration is enabled.
        acme_profile : str, optional
            The profile to use when issuing certificates via ACMEv2. Defaults to the CA_DEFAULT_PROFILE.
        acme_requires_contact : bool, optional
            Set to ``False`` if you do not want to force clients to register with an email address.
        openssh_ca : bool, optional
            Set to ``True`` if you want to use this to use this CA for signing OpenSSH certs.
        sign_certificate_policies : :py:class:`~cg:cryptography.x509.Extension`, optional
            Add the given Certificate Policies extension when signing certificates.
        ocsp_responder_key_validity : int, optional
            How long (in days) OCSP responder keys should be valid.
        ocsp_response_validity : int, optional
            How long (in seconds) OCSP responses should be valid.

        Raises
        ------

        ValueError
            For various cases of wrong input data (e.g. ``key_size`` not being the power of two).
        PermissionError
            If the private key file cannot be written to disk.
        """
        # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,too-many-statements
        # NOTE: Already verified by KeySizeAction, so these checks are only for when the Python API is used
        #       directly. generate_private_key() invokes this again, but we here to avoid sending a signal.

        if extensions is None:
            extensions = []
        else:
            extensions = list(extensions)  # cast extensions to list if set (so that we can extend later)

            # check type of values to provide better errors
            for extension in extensions:
                if isinstance(extension, x509.Extension) is False:
                    raise ValueError(f"Cannot add extension of type {type(extension).__name__}")

        key_size, elliptic_curve = validate_private_key_parameters(key_type, key_size, elliptic_curve)
        algorithm = validate_public_key_parameters(key_type, algorithm)

        expires = parse_expires(expires)

        if openssh_ca and parent:
            raise ValueError("OpenSSH does not support intermediate authorities")

        # Append OpenSSH extensions if an OpenSSH CA was requested
        if openssh_ca:
            extensions.extend([SshHostCaExtension(), SshUserCaExtension()])

        extensions_dict = {ext.oid: ext for ext in extensions}

        if ExtensionOID.KEY_USAGE not in extensions_dict:
            extensions_dict[ExtensionOID.KEY_USAGE] = x509.Extension(
                oid=ExtensionOID.KEY_USAGE,
                critical=constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.KEY_USAGE],
                value=self.model.DEFAULT_KEY_USAGE,
            )

        if crl_url is None:
            crl_url = []

        serial = x509.random_serial_number()

        if default_hostname is None:
            default_hostname = ca_settings.CA_DEFAULT_HOSTNAME

        if acme_profile is None:
            acme_profile = ca_settings.CA_DEFAULT_PROFILE
        elif acme_profile not in ca_settings.CA_PROFILES:
            raise ValueError(f"{acme_profile}: Profile is not defined.")

        if parent:
            signer_serial = parent.pub.loaded.serial_number
        else:
            signer_serial = serial

        context = self._get_formatting_context(serial, signer_serial)

        # If there is a default hostname, use it to compute some URLs from that
        if isinstance(default_hostname, str) and default_hostname:
            default_hostname = validate_hostname(default_hostname, allow_port=True)

            # Set OCSP urls
            if not ocsp_url:
                ocsp_path = reverse("django_ca:ocsp-cert-post", kwargs={"serial": context["SERIAL_HEX"]})
                ocsp_url = f"http://{default_hostname}{ocsp_path}"
            if not issuer_url:
                issuer_url = f"http://{default_hostname}/{context['CA_ISSUER_PATH']}"
            if not crl_url:
                crl_path = reverse("django_ca:crl", kwargs={"serial": context["SERIAL_HEX"]})
                crl_url = [f"http://{default_hostname}{crl_path}"]

            # Add extensions that make sense only for intermediate CAs
            if parent:
                self._handle_authority_information_access(default_hostname, extensions_dict)
                self._handle_crl_distribution_point(default_hostname, extensions_dict)

        # Format extension values
        format_extensions(extensions_dict, context)

        # Cast extensions_dict back to list, so that signal handler receives the same type as the method
        # itself. This has the added bonus of signal handlers being able to influence the extension order.
        extensions = list(extensions_dict.values())

        pre_create_ca.send(
            sender=self.model,
            name=name,
            key_size=key_size,
            key_type=key_type,
            algorithm=algorithm,
            expires=expires,
            parent=parent,
            subject=subject,
            path_length=path_length,
            issuer_url=issuer_url,
            issuer_alt_name=issuer_alt_name,
            crl_url=crl_url,
            ocsp_url=ocsp_url,
            password=password,
            parent_password=parent_password,
            extensions=extensions,
            caa=caa,
            website=website,
            terms_of_service=terms_of_service,
            acme_enabled=acme_enabled,
            acme_registration=acme_registration,
            acme_profile=acme_profile,
            acme_requires_contact=acme_requires_contact,
            sign_certificate_policies=sign_certificate_policies,
            ocsp_responder_key_validity=ocsp_responder_key_validity,
            ocsp_response_validity=ocsp_response_validity,
        )

        private_key = generate_private_key(key_size, key_type, elliptic_curve)
        public_key = private_key.public_key()

        builder = get_cert_builder(expires, serial=serial)
        builder = builder.public_key(public_key)
        builder = builder.subject_name(subject)
        builder = builder.add_extension(
            x509.BasicConstraints(ca=True, path_length=path_length), critical=True
        )

        subject_key_id = x509.SubjectKeyIdentifier.from_public_key(public_key)
        builder = builder.add_extension(subject_key_id, critical=False)

        if parent is None:
            builder = builder.issuer_name(subject)
            private_sign_key = private_key
            aki = x509.AuthorityKeyIdentifier.from_issuer_public_key(public_key)
        else:
            builder = builder.issuer_name(parent.pub.loaded.subject)
            private_sign_key = parent.key(parent_password)
            aki = parent.get_authority_key_identifier()
        builder = builder.add_extension(aki, critical=False)

        for extra_extension in extensions:
            builder = builder.add_extension(extra_extension.value, critical=extra_extension.critical)

        certificate = builder.sign(private_key=private_sign_key, algorithm=algorithm)

        # Normalize extensions for create()
        crl_url = "\n".join(crl_url)

        # Convert arguments for database storage
        serialized_ian = ""
        if issuer_alt_name is not None:
            serialized_ian = ",".join(format_general_name(name) for name in issuer_alt_name.value)

        ca: CertificateAuthority = self.model(
            name=name,
            issuer_url=issuer_url,
            issuer_alt_name=serialized_ian,
            ocsp_url=ocsp_url,
            crl_url=crl_url,
            parent=parent,
            caa_identity=caa,
            website=website,
            terms_of_service=terms_of_service,
            acme_enabled=acme_enabled,
            acme_registration=acme_registration,
            acme_profile=acme_profile,
            acme_requires_contact=acme_requires_contact,
            sign_certificate_policies=sign_certificate_policies,
        )

        # Set fields with a default value
        if ocsp_responder_key_validity is not None:
            ca.ocsp_responder_key_validity = ocsp_responder_key_validity
        if ocsp_response_validity is not None:
            ca.ocsp_response_validity = ocsp_response_validity

        ca.update_certificate(certificate)

        if password is None:
            encryption: serialization.KeySerializationEncryption = serialization.NoEncryption()
        else:
            if isinstance(password, str):
                password = password.encode("utf-8")
            encryption = serialization.BestAvailableEncryption(password)

        pem = private_key.private_bytes(
            encoding=Encoding.PEM, format=PrivateFormat.PKCS8, encryption_algorithm=encryption
        )

        # write private key to file
        safe_serial = ca.serial.replace(":", "")
        path = path / pathlib.PurePath(f"{safe_serial}.key")
        ca.private_key_path = ca_storage.save(str(path), ContentFile(pem))
        ca.save()

        post_create_ca.send(sender=self.model, ca=ca)
        return ca


class CertificateManager(
    CertificateManagerMixin["Certificate", "CertificateQuerySet"], CertificateManagerBase
):
    """Model manager for the Certificate model."""

    if typing.TYPE_CHECKING:
        # See CertificateManagerMixin for description on this branch
        #
        # pylint: disable=missing-function-docstring; just defining stubs here
        def currently_valid(self) -> "CertificateQuerySet":
            ...

        def expired(self) -> "CertificateQuerySet":
            ...

        def not_yet_valid(self) -> "CertificateQuerySet":
            ...

        def revoked(self) -> "CertificateQuerySet":
            ...

    def create_cert(  # pylint: disable=too-many-arguments
        self,
        ca: "CertificateAuthority",
        csr: x509.CertificateSigningRequest,
        profile: Optional[Profile] = None,
        autogenerated: Optional[bool] = None,
        subject: Optional[x509.Name] = None,
        expires: Expires = None,
        algorithm: Optional[AllowedHashTypes] = None,
        extensions: Optional[Iterable[x509.Extension[x509.ExtensionType]]] = None,
        cn_in_san: Optional[bool] = None,
        add_crl_url: Optional[bool] = None,
        add_ocsp_url: Optional[bool] = None,
        add_issuer_url: Optional[bool] = None,
        add_issuer_alternative_name: Optional[bool] = None,
        password: Optional[Union[str, bytes]] = None,
    ) -> "Certificate":
        """Create and sign a new certificate based on the given profile.

        Parameters
        ----------

        ca : :py:class:`~django_ca.models.CertificateAuthority`
            The certificate authority to sign the certificate with.
        csr : :py:class:`~cg:cryptography.x509.CertificateSigningRequest`
            The certificate signing request to use when signing a certificate. Passing a ``str`` or ``bytes``
            is deprecated and will be removed in django-ca 1.20.0.
        profile : :py:class:`~django_ca.profiles.Profile`, optional
            The name of a profile or a manually created :py:class:`~django_ca.profiles.Profile` instance. If
            not given, the profile configured by :ref:`CA_DEFAULT_PROFILE <settings-ca-default-profile>` is
            used.
        autogenerated : bool, optional
            Override the profiles ``autogenerated`` flag.
        subject : :py:class:`~cg:cryptography.x509.Name`, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        expires : int or datetime or timedelta, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        algorithm : :py:class:`~cg:cryptography.hazmat.primitives.hashes.HashAlgorithm`, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        extensions : list or of :py:class:`~cg:cryptography.x509.Extension`
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        cn_in_san : bool, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        cn_in_san : bool, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        add_crl_url : bool, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        add_ocsp_url : bool, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        add_issuer_url : bool, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        add_issuer_alternative_name : bool, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        password : bool, optional
            Passed to :py:func:`Profiles.create_cert() <django_ca.profiles.Profile.create_cert>`.
        """

        # Get the profile object if none was passed
        if profile is None:
            profile = profiles[None]
        elif not isinstance(profile, Profile):
            raise TypeError("profile must be of type django_ca.profiles.Profile.")

        cert = profile.create_cert(
            ca,
            csr,
            subject=subject,
            expires=expires,
            algorithm=algorithm,
            extensions=extensions,
            cn_in_san=cn_in_san,
            add_crl_url=add_crl_url,
            add_ocsp_url=add_ocsp_url,
            add_issuer_url=add_issuer_url,
            add_issuer_alternative_name=add_issuer_alternative_name,
            password=password,
        )

        obj = self.model(ca=ca, csr=LazyCertificateSigningRequest(csr), profile=profile.name)
        obj.update_certificate(cert)
        if autogenerated is None:
            obj.autogenerated = profile.autogenerated
        else:
            obj.autogenerated = autogenerated
        obj.save()

        post_issue_cert.send(sender=self.model, cert=obj)

        return obj


class AcmeAccountManager(AcmeAccountManagerBase):
    """Model manager for :py:class:`~django_ca.models.AcmeAccount`."""

    if typing.TYPE_CHECKING:
        # See CertificateManagerMixin for description on this branch
        #
        # pylint: disable=missing-function-docstring; just defining stubs here

        def viewable(self) -> "AcmeAccountQuerySet":
            ...


class AcmeOrderManager(AcmeOrderManagerBase):
    """Model manager for :py:class:`~django_ca.models.AcmeOrder`."""


class AcmeAuthorizationManager(AcmeAuthorizationManagerBase):
    """Model manager for :py:class:`~django_ca.models.AcmeAuthorization`."""


class AcmeChallengeManager(AcmeChallengeManagerBase):
    """Model manager for :py:class:`~django_ca.models.AcmeChallenge`."""


class AcmeCertificateManager(AcmeCertificateManagerBase):
    """Model manager for :py:class:`~django_ca.models.AcmeCertificate`."""



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to regenerate keys used for OCSP signing.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from datetime import timedelta
from typing import Any, Iterable, Optional

from cryptography.hazmat.primitives.asymmetric import ec

from django.core.management.base import CommandError, CommandParser

from django_ca import ca_settings, constants
from django_ca.management.actions import ExpiresAction
from django_ca.management.base import BaseCommand
from django_ca.models import CertificateAuthority
from django_ca.tasks import generate_ocsp_key, run_task
from django_ca.typehints import AllowedHashTypes, ParsableKeyType
from django_ca.utils import add_colons, validate_private_key_parameters


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = "Regenerate OCSP keys."

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "serials",
            metavar="serial",
            nargs="*",
            help="Generate OCSP keys only for the given CA. If omitted, generate keys for all CAs.",
        )

        parser.add_argument(
            "--expires",
            default=timedelta(days=2),
            action=ExpiresAction,
            help="Sign the certificate for DAYS days (default: %(default)s)",
        )
        parser.add_argument(
            "--force",
            action="store_true",
            default=False,
            help="Force regeneration of OCSP responder certificates.",
        )
        parser.add_argument("--quiet", action="store_true", default=False, help="Do not output warnings.")

        self.add_algorithm(parser)
        private_key_group = parser.add_argument_group("Private key parameters")
        self.add_key_type(
            private_key_group, default=None, default_text="key type of the certificate authority"
        )
        self.add_key_size(private_key_group)
        self.add_elliptic_curve(private_key_group)
        self.add_password(parser)

        self.add_profile(
            parser, 'Override the profile used for generating the certificate. By default, "ocsp" is used.'
        )

    def handle(  # pylint: disable=too-many-arguments
        self,
        serials: Iterable[str],
        profile: Optional[str],
        expires: timedelta,
        algorithm: Optional[AllowedHashTypes],
        key_type: Optional[ParsableKeyType],
        key_size: Optional[int],
        elliptic_curve: Optional[ec.EllipticCurve],
        password: Optional[bytes],
        quiet: bool,
        force: bool,
        **options: Any,
    ) -> None:
        profile = profile or "ocsp"

        # Check if the profile exists. Note that this shouldn't really happen, since valid parameters match
        # existing profiles. The only case is when the user removes the "ocsp" profile, which is the
        # default.
        if profile not in ca_settings.CA_PROFILES:
            raise CommandError(f"{profile}: Undefined profile.")

        if not serials:
            serials = CertificateAuthority.objects.all().order_by("serial").values_list("serial", flat=True)

        for serial in serials:
            serial = serial.replace(":", "").strip().upper()
            hr_serial = add_colons(serial)
            try:
                ca = CertificateAuthority.objects.get(serial=serial)
            except CertificateAuthority.DoesNotExist:
                self.stderr.write(self.style.ERROR(f"{hr_serial}: Unknown CA."))
                continue

            if not ca.key_exists:
                if quiet is False:  # pragma: no branch
                    # NOTE: coverage falsely identifies the above condition to always be false.
                    self.stderr.write(self.style.WARNING(f"{hr_serial}: CA has no private key."))

                continue

            # Get private key parameters for this particular private key
            ca_key_type = key_type
            if ca_key_type is None:
                ca_key_type = ca.key_type

            ca_key_size: Optional[int] = None
            if ca_key_type in ("RSA", "DSA") and key_size is not None:
                ca_key_size = key_size

            ca_elliptic_curve: Optional[ec.EllipticCurve] = None
            if ca_key_type == "EC" and elliptic_curve is not None:
                ca_elliptic_curve = elliptic_curve

            validate_private_key_parameters(ca_key_type, ca_key_size, ca_elliptic_curve)

            algorithm_name: Optional[str] = None
            if algorithm is not None:
                algorithm_name = constants.HASH_ALGORITHM_NAMES[type(algorithm)]

            elliptic_curve_name: Optional[str] = None
            if ca_elliptic_curve is not None:
                elliptic_curve_name = ca_elliptic_curve.name

            run_task(
                generate_ocsp_key,
                ca.serial,
                profile=profile,
                expires=expires.total_seconds(),
                algorithm=algorithm_name,
                key_size=ca_key_size,
                key_type=ca_key_type,
                elliptic_curve=elliptic_curve_name,
                password=password,
                force=force,
            )



# Generated by Django 4.2.3 on 2023-07-30 12:48

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0032_certificateauthority_ocsp_responder_key_validity_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="certificate",
            options={
                "permissions": (
                    ("revoke_certificate", "Can revoke a certificate"),
                    ("sign_certificate", "Can sign a certificate"),
                )
            },
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Module to parse serialized extensions into cryptography objects."""

from typing import Iterable, Iterator, List, Optional, Union

from cryptography import x509
from cryptography.x509.oid import AuthorityInformationAccessOID

from django_ca.constants import (
    EXTENDED_KEY_USAGE_NAMES,
    EXTENSION_DEFAULT_CRITICAL,
    KEY_USAGE_NAMES,
    TLS_FEATURE_NAMES,
)
from django_ca.typehints import (
    ParsableAuthorityInformationAccess,
    ParsableAuthorityKeyIdentifier,
    ParsableBasicConstraints,
    ParsableDistributionPoint,
    ParsableExtension,
    ParsableGeneralNameList,
    ParsableNameConstraints,
    ParsableNoticeReference,
    ParsablePolicyConstraints,
    ParsablePolicyInformation,
    ParsableSubjectKeyIdentifier,
    ParsableUserNotice,
)
from django_ca.utils import hex_to_bytes, parse_general_name, x509_relative_name

##########################################
# Parsers for sub-elements of extensions #
##########################################


def _parse_path_length(value: Optional[Union[int, str]]) -> Optional[int]:
    """Parse `value` as path length (either an int, a str of an int or None)."""
    if value is not None:
        return int(value)
    return value


def _parse_notice_reference(
    value: Optional[Union[x509.NoticeReference, ParsableNoticeReference]],
) -> Optional[x509.NoticeReference]:
    if not value:
        return None
    if isinstance(value, x509.NoticeReference):
        return value

    return x509.NoticeReference(
        organization=value.get("organization"), notice_numbers=value["notice_numbers"]
    )


def _parse_user_notice(value: ParsableUserNotice) -> x509.UserNotice:
    notice_reference = _parse_notice_reference(value.get("notice_reference"))
    return x509.UserNotice(notice_reference=notice_reference, explicit_text=value.get("explicit_text"))


def _parse_policy_qualifiers(
    value: Optional[Iterable[Union[str, x509.UserNotice, ParsableUserNotice]]]
) -> Optional[List[Union[str, x509.UserNotice]]]:
    if not value:
        return None

    qualifiers: List[Union[str, x509.UserNotice]] = []
    for qual in value:
        if isinstance(qual, str):
            qualifiers.append(qual)
        elif isinstance(qual, x509.UserNotice):
            qualifiers.append(qual)
        else:
            qualifiers.append(_parse_user_notice(qual))
    return qualifiers


def _parse_reason(reason: Union[str, x509.ReasonFlags]) -> x509.ReasonFlags:
    if isinstance(reason, str):
        return x509.ReasonFlags[reason]
    return reason


def _parse_distribution_points(
    value: Iterable[Union[x509.DistributionPoint, ParsableDistributionPoint]]
) -> Iterator[x509.DistributionPoint]:
    for dpoint in value:
        if isinstance(dpoint, x509.DistributionPoint):
            yield dpoint
        else:
            full_name = relative_name = reasons = crl_issuer = None

            unparsed_full_name = dpoint.get("full_name")
            if unparsed_full_name is not None:
                full_name = [parse_general_name(name) for name in unparsed_full_name]

            if dpoint.get("relative_name"):
                if isinstance(dpoint["relative_name"], x509.RelativeDistinguishedName):
                    relative_name = dpoint["relative_name"]
                else:
                    relative_name = x509_relative_name(dpoint["relative_name"])

            if dpoint.get("crl_issuer"):
                crl_issuer = [parse_general_name(name) for name in dpoint["crl_issuer"]]

            if dpoint.get("reasons"):
                reasons = frozenset(_parse_reason(r) for r in dpoint["reasons"])

            yield x509.DistributionPoint(
                full_name=full_name, relative_name=relative_name, reasons=reasons, crl_issuer=crl_issuer
            )


#####################
# Extensions parsers#
#####################
def _parse_authority_key_identifier(value: ParsableAuthorityKeyIdentifier) -> x509.AuthorityKeyIdentifier:
    key_id = issuer = serial_number = None

    if isinstance(value, (bytes, str)):
        key_id = value
    else:  # dict
        key_id = value.get("key_identifier")

        if value.get("authority_cert_issuer", []):
            issuer = [parse_general_name(name) for name in value["authority_cert_issuer"]]

        serial_number = value.get("authority_cert_serial_number")
        if serial_number is not None and not isinstance(serial_number, int):
            serial_number = int(serial_number)

    if isinstance(key_id, str):
        key_id = hex_to_bytes(key_id)

    return x509.AuthorityKeyIdentifier(
        key_identifier=key_id, authority_cert_issuer=issuer, authority_cert_serial_number=serial_number
    )


def _parse_authority_information_access(
    value: ParsableAuthorityInformationAccess,
) -> x509.AuthorityInformationAccess:
    access_descriptions: List[x509.AccessDescription] = []
    issuers: Optional[ParsableGeneralNameList] = value.get("issuers")
    if issuers is None:
        issuers = []

    for name in issuers:
        access_descriptions.append(
            x509.AccessDescription(
                access_method=AuthorityInformationAccessOID.CA_ISSUERS,
                access_location=parse_general_name(name),
            )
        )
    ocsp: Optional[ParsableGeneralNameList] = value.get("ocsp")
    if ocsp is None:
        ocsp = []
    for name in ocsp:
        access_descriptions.append(
            x509.AccessDescription(
                access_method=AuthorityInformationAccessOID.OCSP,
                access_location=parse_general_name(name),
            )
        )

    return x509.AuthorityInformationAccess(descriptions=access_descriptions)


def _parse_basic_constraints(value: ParsableBasicConstraints) -> x509.BasicConstraints:
    ca = bool(value.get("ca", False))
    path_length = None
    if ca is True:
        path_length = _parse_path_length(value.get("path_length"))
    return x509.BasicConstraints(ca=value["ca"], path_length=path_length)


def _parse_certificate_policies(
    value: Iterable[Union[x509.PolicyInformation, ParsablePolicyInformation]]
) -> x509.CertificatePolicies:
    policies: List[x509.PolicyInformation] = []
    for pol in value:
        if isinstance(pol, x509.PolicyInformation):
            policies.append(pol)
            continue

        policy_identifier = pol.get("policy_identifier")
        if isinstance(policy_identifier, str):
            parsed_policy_identifier = x509.ObjectIdentifier(policy_identifier)
        else:
            # TYPE NOTE: bug in current typehints module: policy_identifier cannot be None
            parsed_policy_identifier = policy_identifier  # type: ignore[assignment]

        qualifiers = _parse_policy_qualifiers(pol.get("policy_qualifiers"))
        policies.append(
            x509.PolicyInformation(policy_identifier=parsed_policy_identifier, policy_qualifiers=qualifiers)
        )

    return x509.CertificatePolicies(policies)


def _parse_crl_distribution_points(
    value: Iterable[Union[x509.DistributionPoint, ParsableDistributionPoint]]
) -> x509.CRLDistributionPoints:
    return x509.CRLDistributionPoints(distribution_points=_parse_distribution_points(value))


def _parse_extended_key_usage(value: Iterable[Union[str, x509.ObjectIdentifier]]) -> x509.ExtendedKeyUsage:
    mapping = {v: k for k, v in EXTENDED_KEY_USAGE_NAMES.items()}
    usages: List[x509.ObjectIdentifier] = []
    for unparsed in value:
        if isinstance(unparsed, str):
            try:
                usages.append(mapping[unparsed])
            except KeyError:
                usages.append(x509.ObjectIdentifier(unparsed))
        else:
            usages.append(unparsed)

    return x509.ExtendedKeyUsage(
        usages=sorted(usages, key=lambda u: EXTENDED_KEY_USAGE_NAMES.get(u, u.dotted_string))
    )


def _parse_freshest_crl(
    value: Iterable[Union[x509.DistributionPoint, ParsableDistributionPoint]]
) -> x509.FreshestCRL:
    return x509.FreshestCRL(distribution_points=_parse_distribution_points(value))


def _parse_key_usage(value: Iterator[str]) -> x509.KeyUsage:
    kwargs = {k: k in value or v in value for k, v in KEY_USAGE_NAMES.items()}
    return x509.KeyUsage(**kwargs)


def _parse_name_constraints(value: ParsableNameConstraints) -> x509.NameConstraints:
    permitted = value.get("permitted")
    if not permitted:
        permitted_subtrees = None
    else:
        permitted_subtrees = [parse_general_name(name) for name in permitted]

    excluded = value.get("excluded")
    if not excluded:
        excluded_subtrees = None
    else:
        excluded_subtrees = [parse_general_name(name) for name in excluded]
    return x509.NameConstraints(permitted_subtrees=permitted_subtrees, excluded_subtrees=excluded_subtrees)


def _parse_policy_constraints(value: ParsablePolicyConstraints) -> x509.PolicyConstraints:
    return x509.PolicyConstraints(
        require_explicit_policy=value.get("require_explicit_policy"),
        inhibit_policy_mapping=value.get("inhibit_policy_mapping"),
    )


def _parse_subject_key_identifier(value: ParsableSubjectKeyIdentifier) -> x509.SubjectKeyIdentifier:
    if isinstance(value, x509.SubjectKeyIdentifier):
        return value
    if isinstance(value, str):
        value = hex_to_bytes(value)
    return x509.SubjectKeyIdentifier(digest=value)


def _parse_tls_feature(value: Iterable[Union[x509.TLSFeatureType, str]]) -> x509.TLSFeature:
    features: List[x509.TLSFeatureType] = []
    for feature in value:
        if isinstance(feature, str):
            feature = TLS_FEATURE_NAMES[feature]
        features.append(feature)

    features = sorted(features, key=lambda f: f.name)
    return x509.TLSFeature(features=features)


def parse_extension(  # pylint: disable=too-many-branches  # there's just many extensions
    key: str, value: Union[x509.Extension[x509.ExtensionType], x509.ExtensionType, ParsableExtension]
) -> x509.Extension[x509.ExtensionType]:
    """Parse a serialized extension into a cryptography object.

    This function is used by :doc:`profiles` to parse configured extensions into standard cryptography
    extensions. If you need to parse a similar object, use this function.

    The `value` is usually a ``dict`` as described in profiles but for convenience, may also be a
    :py:class:`~cg:cryptography.x509.Extension`, in which case the extension is returned unchanged. If you
    pass a :py:class:`~cg:cryptography.x509.ExtensionType`, an extension with the default critical value is
    returned.

    >>> parse_extension("key_usage", {'critical': True, 'value': ['keyCertSign']})  # doctest: +ELLIPSIS
    <Extension(..., critical=True, value=<KeyUsage(... key_cert_sign=True, ...)>)>

    Parameters
    ----------
    key : str
        The `key` is the extension key used in the dictionary to name the extension, it must match one of the
        keys in :py:data:`~django_ca.constants.EXTENSION_KEYS`.
    value : dict, |ExtensionType| or |Extension|
        The value that describes the extension. See :doc:`/profiles` for more information.
    """
    if isinstance(value, x509.Extension):
        return value

    if isinstance(value, x509.ExtensionType):
        return x509.Extension(oid=value.oid, critical=EXTENSION_DEFAULT_CRITICAL[value.oid], value=value)

    if key == "authority_key_identifier":
        parsed: x509.ExtensionType = _parse_authority_key_identifier(value["value"])
    elif key == "authority_information_access":
        parsed = _parse_authority_information_access(value["value"])
    elif key == "basic_constraints":
        parsed = _parse_basic_constraints(value["value"])
    elif key == "certificate_policies":
        parsed = _parse_certificate_policies(value["value"])
    elif key == "crl_distribution_points":
        parsed = _parse_crl_distribution_points(value["value"])
    elif key == "freshest_crl":
        parsed = _parse_freshest_crl(value["value"])
    elif key == "extended_key_usage":
        parsed = _parse_extended_key_usage(value["value"])
    elif key == "issuer_alternative_name":
        parsed = x509.IssuerAlternativeName([parse_general_name(name) for name in value["value"]])
    elif key == "key_usage":
        parsed = _parse_key_usage(value["value"])
    elif key == "name_constraints":
        parsed = _parse_name_constraints(value["value"])
    elif key == "ocsp_no_check":
        parsed = x509.OCSPNoCheck()
    elif key == "inhibit_any_policy":
        parsed = x509.InhibitAnyPolicy(skip_certs=value["value"])
    elif key == "policy_constraints":
        parsed = _parse_policy_constraints(value["value"])
    elif key == "precert_poison":
        parsed = x509.PrecertPoison()
    elif key == "subject_alternative_name":
        parsed = x509.SubjectAlternativeName([parse_general_name(name) for name in value["value"]])
    elif key == "subject_key_identifier":
        parsed = _parse_subject_key_identifier(value["value"])
    elif key == "tls_feature":
        parsed = _parse_tls_feature(value["value"])
    elif key in ("precertificate_signed_certificate_timestamps", "signed_certificate_timestamps"):
        # https://github.com/pyca/cryptography/issues/7824
        raise ValueError(f"{key}: Cannot parse extensions of this type.")
    else:
        raise ValueError(f"{key}: Unknown extension key.")

    critical = value.get("critical", EXTENSION_DEFAULT_CRITICAL[parsed.oid])
    return x509.Extension(oid=parsed.oid, critical=critical, value=parsed)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Extension classes wrapping various X.509 extensions.

The classes in this module wrap cryptography extensions, but allow adding/removing values, creating extensions
in a more pythonic manner and provide access functions."""

from cryptography import x509
from cryptography.hazmat._oid import _OID_NAMES as OID_NAMES

from django_ca.constants import EXTENSION_NAMES
from django_ca.extensions.parse import parse_extension
from django_ca.extensions.serialize import serialize_extension
from django_ca.extensions.text import extension_as_text

#: Tuple of extensions that can be set when creating a new certificate
CERTIFICATE_EXTENSIONS = tuple(
    sorted(
        [
            "authority_information_access",
            "certificate_policies",
            "crl_distribution_points",
            "extended_key_usage",
            "freshest_crl",
            "issuer_alternative_name",
            "key_usage",
            "ocsp_no_check",
            "tls_feature",
        ]
    )
)


def get_extension_name(oid: x509.ObjectIdentifier) -> str:
    """Function to get the name of an extension from the extensions OID.

    >>> get_extension_name(ExtensionOID.BASIC_CONSTRAINTS)
    'Basic Constraints'
    >>> get_extension_name(x509.ObjectIdentifier("1.2.3"))
    'Unknown extension (1.2.3)'

    """

    if oid in EXTENSION_NAMES:
        return EXTENSION_NAMES[oid]

    return OID_NAMES.get(oid, f"Unknown extension ({oid.dotted_string})")


__all__ = [
    "extension_as_text",
    "get_extension_name",
    "parse_extension",
    "serialize_extension",
]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""django-ca model fields.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-model-fields/
"""

import abc
import json
import typing
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from cryptography import x509
from cryptography.hazmat.primitives.serialization import Encoding

import django
from django import forms
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from django_ca import constants
from django_ca.extensions import parse_extension, serialize_extension
from django_ca.fields import CertificateSigningRequestField as CertificateSigningRequestFormField
from django_ca.typehints import JSON, ExtensionTypeTypeVar, SerializedExtension

DecodableCertificate = Union[str, bytes, x509.Certificate]
DecodableCertificateSigningRequest = Union[str, bytes, x509.CertificateSigningRequest]
LoadedTypeVar = typing.TypeVar("LoadedTypeVar", x509.CertificateSigningRequest, x509.Certificate)
DecodableTypeVar = typing.TypeVar(
    "DecodableTypeVar", DecodableCertificate, DecodableCertificateSigningRequest
)
# TYPE NOTE: mypy does not support using generics as bound ("bound=LazyField[DecodableTypeVar, ...]").
#   If we list concrete subclasses instead, typing functions that take DecodableTypeVar and return a
#   WrapperTypeVar instance becomes difficult, as mypy has no way of knowing that DecodableTypeVar argument
#   matches the type in required by the constructor of WrapperTypeVar.
WrapperTypeVar = typing.TypeVar("WrapperTypeVar", bound="LazyField")  # type: ignore[type-arg]

# django-stubs models.Field subclasses generic, so we need  a different base when type checking.
if typing.TYPE_CHECKING:

    class LazyBinaryFieldBase(
        models.BinaryField[Union[DecodableTypeVar, WrapperTypeVar], WrapperTypeVar],
        typing.Generic[DecodableTypeVar, WrapperTypeVar],
    ):
        # pylint: disable=missing-class-docstring
        pass

else:

    class LazyBinaryFieldBase(models.BinaryField, typing.Generic[DecodableTypeVar, WrapperTypeVar]):
        # pylint: disable=missing-class-docstring
        pass


class LazyField(typing.Generic[LoadedTypeVar, DecodableTypeVar], metaclass=abc.ABCMeta):
    """Abstract base class for lazy field values.

    Subclasses of this class can be used by *binary* fields to load a cryptography value when first accessed.
    """

    _bytes: bytes
    _loaded: Optional[LoadedTypeVar] = None
    _pem_token: typing.ClassVar[bytes]
    _type: Type[LoadedTypeVar]

    def __init__(self, value: DecodableTypeVar) -> None:
        """Constructor must accept a decodable type var."""
        if isinstance(value, bytes):  # SQLite passes bytes
            if value.startswith(self._pem_token):
                self._loaded = self.load_pem(value)
                value = self._loaded.public_bytes(Encoding.DER)

            self._bytes = value
        elif isinstance(value, bytearray):
            self._bytes = bytes(value)
        elif isinstance(value, memoryview):  # PostgreSQL driver passes memoryview
            self._bytes = value.tobytes()
        elif isinstance(value, str):
            self._loaded = self.load_pem(value.encode())
            self._bytes = self._loaded.public_bytes(Encoding.DER)
        elif isinstance(value, self._type):
            self._loaded = value
            self._bytes = self._loaded.public_bytes(Encoding.DER)
        else:
            raise ValueError(f"{value}: Could not parse {self._type.__name__}")

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self._bytes == other._bytes

    def __repr__(self) -> str:
        name = self.__class__.__name__
        value = self.loaded.subject.rfc4514_string()
        return f"<{name}: {value}>"

    @abc.abstractmethod
    def load_pem(self, value: bytes) -> LoadedTypeVar:
        """Return the loaded value from the given bytes."""

    @property
    @abc.abstractmethod
    def loaded(self) -> LoadedTypeVar:
        """The stored value parsed into a cryptography object."""

    def encode(self, encoding: Encoding) -> bytes:
        """
        Parameters
        ----------

        encoding : attr of :py:class:`~cg:cryptography.hazmat.primitives.serialization.Encoding`, optional
            The format to return, defaults to ``Encoding.PEM``.
        """
        if encoding == Encoding.DER:
            return self._bytes
        return self.loaded.public_bytes(encoding)

    @property
    def der(self) -> bytes:
        """This field in its raw DER representation."""
        return self._bytes

    @property
    def pem(self) -> str:
        """This CSR as str-encoded PEM."""
        return self.loaded.public_bytes(Encoding.PEM).decode()


class LazyCertificateSigningRequest(
    LazyField[x509.CertificateSigningRequest, DecodableCertificateSigningRequest]
):
    """A lazy field for a :py:class:`~cg:cryptography.x509.CertificateSigningRequest."""

    _pem_token = b"-----BEGIN CERTIFICATE REQUEST-----"
    _type = x509.CertificateSigningRequest

    def load_pem(self, value: bytes) -> x509.CertificateSigningRequest:
        return x509.load_pem_x509_csr(value)

    @property
    def loaded(self) -> x509.CertificateSigningRequest:
        """This CSR as :py:class:`cg:cryptography.x509.CertificateSigningRequest`."""
        if self._loaded is None:
            self._loaded = x509.load_der_x509_csr(self._bytes)
        return self._loaded


class LazyCertificate(LazyField[x509.Certificate, DecodableCertificate]):
    """A lazy field for a :py:class:`~cg:cryptography.x509.Certificate."""

    _pem_token = b"-----BEGIN CERTIFICATE-----"
    _type = x509.Certificate

    def load_pem(self, value: bytes) -> x509.Certificate:
        return x509.load_pem_x509_certificate(value)

    @property
    def loaded(self) -> x509.Certificate:
        """This CSR as :py:class:`cg:cryptography.x509.CertificateSigningRequest`."""
        if self._loaded is None:
            self._loaded = x509.load_der_x509_certificate(self._bytes)
        return self._loaded


class LazyBinaryField(
    LazyBinaryFieldBase[DecodableTypeVar, WrapperTypeVar], typing.Generic[DecodableTypeVar, WrapperTypeVar]
):
    """Base class for binary fields that parse the value when first used."""

    formfield_class: Type[forms.Field]
    wrapper: Type[WrapperTypeVar]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs.setdefault("editable", True)
        super().__init__(*args, **kwargs)

    def deconstruct(self) -> Tuple[str, str, List[str], Dict[str, str]]:
        """Used in migrations."""
        name, path, args, kwargs = super().deconstruct()

        # COVERAGE NOTE: parent field sets this to False, so we override the default. But we don't actually
        #   need a non-editable field, so we never manually set this to False.
        if self.editable is True:  # pragma: no branch
            del kwargs["editable"]

        return name, path, args, kwargs

    def from_db_value(  # pylint: disable=unused-argument
        self, value: Optional[bytes], expression: Any, condition: Any
    ) -> Optional[WrapperTypeVar]:
        """Called when data is loaded from the database.

        This is called when

        * An object is loaded from the database (``Certificate.objects.get()``, ...)
        * A queryset is used (*not* just creating it -> QS are lazy)
        """
        if value is None:
            return None
        return self.wrapper(value)

    def get_prep_value(
        self,
        value: Optional[Union[WrapperTypeVar, DecodableTypeVar]],
    ) -> Optional[bytes]:
        """Get the raw database value.

        This is called when

        * ``Certificate.save()`` is called
        * Also called for queryset methods when using a field of this type (e.g. ``.filter(csr=...)``).
        """
        if not value:
            return None
        if isinstance(value, self.wrapper):
            return value.der
        return self.wrapper(value).der

    def formfield(
        self,
        form_class: Optional[Type[forms.Field]] = None,
        choices_form_class: Optional[Type[forms.Field]] = None,
        **kwargs: Any,
    ) -> forms.Field:
        # COVERAGE NOTE: not None e.g. for ModelForm which defines a form field, but we never do that.
        if form_class is None:  # pragma: no branch
            form_class = self.formfield_class
        # TYPEHINT NOTE: return value of parent is typed to any.
        return super().formfield(form_class, choices_form_class, **kwargs)  # type: ignore[no-any-return]

    def to_python(
        self,
        value: Optional[Union[WrapperTypeVar, DecodableTypeVar]],
    ) -> Optional[WrapperTypeVar]:
        """Called during deserialization and during Certificate.full_clean().

        Note that this function is **not** called if the field value is ``None`` or ``b""``. It is however
        called with ``""`` (the empty string).
        """
        if not value:
            return None
        if isinstance(value, self.wrapper):
            return value
        return self.wrapper(value)


class CertificateSigningRequestField(
    LazyBinaryField[DecodableCertificateSigningRequest, LazyCertificateSigningRequest]
):
    """Django model field for CSRs."""

    formfield_class = CertificateSigningRequestFormField
    wrapper = LazyCertificateSigningRequest


class CertificateField(LazyBinaryField[DecodableCertificate, LazyCertificate]):
    """Django model field for Certificates."""

    # NOTE: Since certificates are never submitted in a form, there never is an active form field for this
    #       field, and formfield() is never called for this class. Thus, it's okay to use the wrong class.
    formfield_class = CertificateSigningRequestFormField
    wrapper = LazyCertificate


class ExtensionField(models.JSONField, typing.Generic[ExtensionTypeTypeVar]):
    """Base class for fields storing a `x509.Extension` class.

    Since the docs are a bit confusing, here is how the methods are called in some scenarios

    full_clean():
    1. to_python() (with the value stored in the instance)
    2. validate() (always an ext, thanks to to_python()

    save():
    1. get_prep_value()

    load():
    1. from_db_value() (with `value` being a string)
    """

    extension_class: Type[ExtensionTypeTypeVar]
    default_error_messages = {
        "unparsable-extension": _("The value cannot be parsed to an extension."),
        "invalid-type": _("%(value)s: Not a cryptography.x509.Extension class."),
        "invalid-extension-type": _("Expected an instance of %(extension_class)s."),
    }

    if typing.TYPE_CHECKING:

        def __get__(  # type: ignore[override]
            self, instance: Any, owner: Any
        ) -> Optional[x509.Extension[ExtensionTypeTypeVar]]:
            ...

        def __set__(
            self,
            instance: Any,
            value: Optional[Union[x509.Extension[x509.ExtensionType], SerializedExtension]],
        ) -> None:
            ...

    @property
    def extension_key(self) -> str:
        """The extension key for the handled extension."""
        return constants.EXTENSION_KEYS[self.extension_class.oid]

    def from_db_value(
        self, value: Any, expression: Any, connection: Any
    ) -> Optional[x509.Extension[ExtensionTypeTypeVar]]:
        """Convert the value loaded from the database to a cryptography extension."""
        if value is None:
            return value

        # TYPE NOTE: django-stubs seems to not have the function in the super-class
        parsed_json: JSON = super().from_db_value(value, expression, connection)  # type: ignore[misc]

        return parse_extension(self.extension_key, parsed_json)  # type: ignore[return-value,arg-type]

    def to_python(self, value: Any) -> Optional[x509.Extension[ExtensionTypeTypeVar]]:
        if isinstance(value, x509.Extension):
            return value

        # COVERAGE NOTE: Despite extensive tests, this method never seems to be called with `value=None`. The
        # docs however strongly recommend that we handle this case, hence the block below.
        if value is None:  # pragma: no cover
            return value

        try:
            return parse_extension(self.extension_key, value)  # type: ignore
        except Exception as ex:
            raise ValidationError(
                self.error_messages["unparsable-extension"],
                code="unparsable-extension",
                params={"value": value},
            ) from ex

    def get_prep_value(self, value: Any) -> Optional[SerializedExtension]:
        """Prepare the value so that it can be stored in the database.

        This function is invoked during ``save()``. `value` may be the cryptography extension value (in
        particular, if ``full_clean()`` was called before) or the serialized extension.
        """
        if value is None:  # pragma: no cover  # this happens during migrations
            return value
        if isinstance(value, dict):
            # Run to_python() to make sure that we can deserialize the extension again when loading.
            # Otherwise, we could just pass any dict and it would work.
            self.to_python(value)

            if django.VERSION < (4, 2):  # pragma: django<4.2 branch
                return json.dumps(value)  # type: ignore[return-value]
            return value  # type: ignore[return-value]  # pragma: django>=4.2 branch
        if not isinstance(value, x509.Extension):
            raise ValidationError(
                self.error_messages["invalid-type"],
                code="invalid-type",
                params={"value": value},
            )
        if not isinstance(value.value, self.extension_class):
            raise ValidationError(
                self.error_messages["invalid-extension-type"],
                code="invalid-extension-type",
                params={"extension_class": self.extension_class.__name__},
            )

        serialized = serialize_extension(value)
        if django.VERSION < (4, 2):  # pragma: django<4.2 branch
            return json.dumps(serialized)  # type: ignore[return-value]
        return serialized  # pragma: django>=4.2 branch

    def validate(self, value: x509.Extension[ExtensionTypeTypeVar], model_instance: Any) -> None:
        """Handle field-specific validation.

        This method is called during full_clean(), but *after* `to_python()` is called. We can thus expect
        `value` to be the appropriate extension already, and we just double-check.

        Note that we *have to* override the function of the parent class, as JSONField.validate() validates
        that the value is JSON serializable (which the extension of course isn't).
        """
        if not isinstance(value, x509.Extension):  # pragma: no cover
            raise ValidationError(
                self.error_messages["invalid-type"],
                code="invalid-type",
                params={"value": value},
            )
        if not isinstance(value.value, self.extension_class):  # pragma: no cover
            raise ValidationError(
                self.error_messages["invalid-extension-type"],
                code="invalid-type",
                params={"extension_class": self.extension_class.__name__},
            )


class CertificatePoliciesField(ExtensionField[x509.CertificatePolicies]):
    """Field storing a :py:class:`~cg:cryptography.x509.CertificatePolicies`-based extension."""

    description = _("A Certificate Policies extension object.")
    extension_class = x509.CertificatePolicies



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Module collecting support classes for OpenSSH CA support."""

from django_ca.openssh.constants import OpenSshCertificateType
from django_ca.openssh.extensions import SshHostCaExtension, SshHostCaType, SshUserCaExtension, SshUserCaType

__all__ = (
    "OpenSshCertificateType",
    "SshHostCaType",
    "SshHostCaExtension",
    "SshUserCaType",
    "SshUserCaExtension",
)



# Generated by Django 4.2.3 on 2023-07-15 16:29

from django.db import migrations
import django_ca.modelfields


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0030_alter_certificateauthority_issuer_alt_name_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="certificateauthority",
            name="sign_certificate_policies",
            field=django_ca.modelfields.CertificatePoliciesField(
                blank=True,
                default=None,
                help_text="Add a Certificate Policies extension when signing certificates.",
                null=True,
                verbose_name="Certificate Policies",
            ),
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""URL configuration for this project."""
from typing import List, Union

from django.conf import settings
from django.urls import URLPattern, URLResolver, path, register_converter

from django_ca import ca_settings, converters, views

app_name = "django_ca"

register_converter(converters.AcmeSlugConverter, "acme")
register_converter(converters.Base64Converter, "base64")
register_converter(converters.HexConverter, "hex")
register_converter(converters.SerialConverter, "serial")

urlpatterns: List[Union[URLResolver, URLPattern]] = [
    path("issuer/<hex:serial>.der", views.GenericCAIssuersView.as_view(), name="issuer"),
    path("ocsp/<hex:serial>/cert/", views.GenericOCSPView.as_view(), name="ocsp-cert-post"),
    path("ocsp/<hex:serial>/cert/<base64:data>", views.GenericOCSPView.as_view(), name="ocsp-cert-get"),
    path("ocsp/<hex:serial>/ca/", views.GenericOCSPView.as_view(ca_ocsp=True), name="ocsp-ca-post"),
    path(
        "ocsp/<hex:serial>/ca/<base64:data>", views.GenericOCSPView.as_view(ca_ocsp=True), name="ocsp-ca-get"
    ),
    path("crl/<hex:serial>/", views.CertificateRevocationListView.as_view(), name="crl"),
    path("crl/ca/<hex:serial>/", views.CertificateRevocationListView.as_view(scope="ca"), name="ca-crl"),
]

if ca_settings.CA_ENABLE_REST_API is True:
    from django_ca.api.endpoints import api

    urlpatterns.append(path("api/", api.urls))


if ca_settings.CA_ENABLE_ACME:
    from django_ca.acme import views as acme_views

    # NOTE: Some functions depend on the fact that ALL ACME urls have a "serial" kwarg
    urlpatterns += [
        path("acme/directory/", acme_views.AcmeDirectory.as_view(), name="acme-directory"),
        path("acme/directory/<serial:serial>/", acme_views.AcmeDirectory.as_view(), name="acme-directory"),
        path("acme/<serial:serial>/new-nonce/", acme_views.AcmeNewNonceView.as_view(), name="acme-new-nonce"),
        path(
            "acme/<serial:serial>/new-account/",
            acme_views.AcmeNewAccountView.as_view(),
            name="acme-new-account",
        ),
        path("acme/<serial:serial>/new-order/", acme_views.AcmeNewOrderView.as_view(), name="acme-new-order"),
        path(
            "acme/<serial:serial>/acct/<acme:slug>/",
            acme_views.AcmeAccountView.as_view(),
            name="acme-account",
        ),
        path(
            "acme/<serial:serial>/acct/<acme:slug>/orders/",
            acme_views.AcmeAccountOrdersView.as_view(),
            name="acme-account-orders",
        ),
        path(
            "acme/<serial:serial>/order/<acme:slug>/", acme_views.AcmeOrderView.as_view(), name="acme-order"
        ),
        path(
            "acme/<serial:serial>/order/<acme:slug>/finalize/",
            acme_views.AcmeOrderFinalizeView.as_view(),
            name="acme-order-finalize",
        ),
        path(
            "acme/<serial:serial>/authz/<acme:slug>/",
            acme_views.AcmeAuthorizationView.as_view(),
            name="acme-authz",
        ),
        path(
            "acme/<serial:serial>/chall/<acme:slug>/",
            acme_views.AcmeChallengeView.as_view(),
            name="acme-challenge",
        ),
        path(
            "acme/<serial:serial>/cert/<acme:slug>/",
            acme_views.AcmeCertificateView.as_view(),
            name="acme-cert",
        ),
        path(
            "acme/<serial:serial>/revoke/",
            acme_views.AcmeCertificateRevocationView.as_view(),
            name="acme-revoke",
        ),
    ]


for name, kwargs in getattr(settings, "CA_OCSP_URLS", {}).items():
    kwargs.setdefault("ca", name)
    urlpatterns += [
        path(f"ocsp/{name}/", views.OCSPView.as_view(**kwargs), name=f"ocsp-post-{name}"),
        path(f"ocsp/{name}/<base64:data>", views.OCSPView.as_view(**kwargs), name=f"ocsp-get-{name}"),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Endpoint implementation for the API."""

from http import HTTPStatus
from typing import List

from ninja import NinjaAPI, Query
from ninja.errors import HttpError

from cryptography import x509

from django.core.exceptions import ValidationError
from django.core.handlers.wsgi import WSGIRequest
from django.http import Http404, HttpResponse

from django_ca import __version__, constants
from django_ca.api.auth import BasicAuth
from django_ca.api.errors import Forbidden
from django_ca.api.schemas import (
    CertificateAuthorityFilterSchema,
    CertificateAuthoritySchema,
    CertificateAuthorityUpdateSchema,
    CertificateFilterSchema,
    CertificateSchema,
    RevokeCertificateSchema,
    SignCertificateSchema,
)
from django_ca.api.utils import get_certificate_authority
from django_ca.extensions import parse_extension
from django_ca.models import Certificate, CertificateAuthority
from django_ca.profiles import profiles
from django_ca.querysets import CertificateAuthorityQuerySet, CertificateQuerySet

api = NinjaAPI(title="django-ca API", version=__version__, urls_namespace="django_ca:api")


@api.exception_handler(Forbidden)
def forbidden(request: WSGIRequest, exc: Exception) -> HttpResponse:  # pylint: disable=unused-argument
    """Add the exception handler for the Forbidden exception."""
    return api.create_response(request, {"detail": "Forbidden"}, status=HTTPStatus.FORBIDDEN)


@api.get(
    "/ca/",
    response=List[CertificateAuthoritySchema],
    auth=BasicAuth("django_ca.view_certificateauthority"),
    summary="List available certificate authorities",
    tags=["Certificate authorities"],
)
def list_certificate_authorities(
    request: WSGIRequest, filters: CertificateAuthorityFilterSchema = Query(...)
) -> CertificateAuthorityQuerySet:
    """Retrieve a list of currently usable certificate authorities."""
    qs = CertificateAuthority.objects.enabled()
    if filters.expired is False:
        qs = qs.valid()
    return qs


@api.get(
    "/ca/{serial:serial}/",
    response=CertificateAuthoritySchema,
    auth=BasicAuth("django_ca.view_certificateauthority"),
    summary="View certificate authority",
    tags=["Certificate authorities"],
)
def view_certificate_authority(request: WSGIRequest, serial: str) -> CertificateAuthority:
    """Retrieve details of the certificate authority with the given serial."""
    return get_certificate_authority(serial, expired=True)  # You can *view* expired CAs


@api.put(
    "/ca/{serial:serial}/",
    response=CertificateAuthoritySchema,
    auth=BasicAuth("django_ca.change_certificateauthority"),
    summary="Update certificate authority",
    tags=["Certificate authorities"],
)
def update_certificate_authority(
    request: WSGIRequest, serial: str, data: CertificateAuthorityUpdateSchema
) -> CertificateAuthority:
    """Update a certificate authority.

    All request body fields are optional, so you can also update only individual fields.
    """
    ca = get_certificate_authority(serial, expired=True)
    for attr, value in data.dict(exclude_unset=True).items():
        setattr(ca, attr, value)

    try:
        ca.full_clean()
    except ValidationError as ex:
        raise HttpError(HTTPStatus.BAD_REQUEST, ex.message_dict) from ex  # type: ignore[arg-type]

    ca.save()
    return ca


@api.post(
    "/ca/{serial:serial}/sign/",
    response=CertificateSchema,
    auth=BasicAuth("django_ca.sign_certificate"),
    summary="Sign a certificate",
    tags=["Certificates"],
)
def sign_certificate(request: WSGIRequest, serial: str, data: SignCertificateSchema) -> Certificate:
    """Sign a certificate.

    The `extensions` value is completely optional and allows you to add additional extensions to the
    certificate. Usually extensions are defined either by the CA or by the named profile.
    """
    csr = x509.load_pem_x509_csr(data.csr.encode())
    ca = get_certificate_authority(serial)
    profile = profiles[data.profile]
    subject = x509.Name.from_rfc4514_string(data.subject)
    algorithm = expires = None
    extensions: List[x509.Extension[x509.ExtensionType]] = []

    if ca.key_exists is False:
        raise HttpError(
            HTTPStatus.BAD_REQUEST,
            "This certificate authority can not be used to sign certificates via the API.",
        )

    if data.algorithm is not None:
        algorithm = constants.HASH_ALGORITHM_TYPES[data.algorithm]()
    if data.expires is not None:
        expires = data.expires

    if data.extensions is not None:
        for extension_key, extension_data in data.extensions:
            if extension_data is not None:
                extensions.append(parse_extension(extension_key, extension_data.dict()))

    cert = Certificate.objects.create_cert(
        ca,
        csr,
        profile=profile,
        autogenerated=data.autogenerated,
        subject=subject,
        expires=expires,
        algorithm=algorithm,
        extensions=extensions,
    )

    return cert


@api.get(
    "/ca/{serial:serial}/certs/",
    response=List[CertificateSchema],
    auth=BasicAuth("django_ca.view_certificate"),
    summary="List certificates",
    tags=["Certificates"],
)
def list_certificates(
    request: WSGIRequest, serial: str, filters: CertificateFilterSchema = Query(...)
) -> CertificateQuerySet:
    """Retrieve certificates signed by the certificate authority named by `serial`."""
    ca = get_certificate_authority(serial, expired=True)  # You can list certificates of expired CAs
    qs = Certificate.objects.filter(ca=ca)

    if filters.expired is False:
        qs = qs.currently_valid()
    if filters.autogenerated is False:
        qs = qs.exclude(autogenerated=True)
    if filters.revoked is False:
        qs = qs.exclude(revoked=True)
    if filters.profile is not None:
        qs = qs.filter(profile=filters.profile)

    return qs


@api.get(
    "/ca/{serial:serial}/certs/{serial:certificate_serial}/",
    response=CertificateSchema,
    auth=BasicAuth("django_ca.view_certificate"),
    summary="View certificate",
    tags=["Certificates"],
)
def view_certificate(request: WSGIRequest, serial: str, certificate_serial: str) -> Certificate:
    """Retrieve details of the certificate with the given certificate serial."""
    ca = get_certificate_authority(serial, expired=True)  # You can view certificates of expired CAs
    return Certificate.objects.get(ca=ca, serial=certificate_serial)


@api.post(
    "/ca/{serial:serial}/revoke/{serial:certificate_serial}/",
    response=CertificateSchema,
    auth=BasicAuth("django_ca.revoke_certificate"),
    summary="Revoke certificate",
    tags=["Certificates"],
)
def revoke_certificate(
    request: WSGIRequest, serial: str, certificate_serial: str, revocation: RevokeCertificateSchema
) -> Certificate:
    """Revoke a certificate with the given serial.

    Both `reason` and `compromised` fields are optional.
    """
    ca = get_certificate_authority(serial)
    try:
        cert = Certificate.objects.currently_valid().get(ca=ca, serial=certificate_serial)
    except Certificate.DoesNotExist as ex:
        raise Http404(f"{certificate_serial}: Certificate not found.") from ex

    if cert.revoked is True:
        raise HttpError(HTTPStatus.BAD_REQUEST, "The certificate is already revoked.")

    cert.revoke(revocation.reason, revocation.compromised)
    return cert



# Generated by Django 4.2.2 on 2023-06-24 10:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0029_certificate_updated_certificateauthority_updated_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="certificateauthority",
            name="issuer_alt_name",
            field=models.CharField(
                blank=True, help_text="URL for your CA.", max_length=255, verbose_name="issuerAltName"
            ),
        ),
        migrations.AlterField(
            model_name="certificateauthority",
            name="issuer_url",
            field=models.URLField(
                blank=True,
                default="",
                help_text="URL to the certificate of this CA (in DER format).",
                verbose_name="Issuer URL",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="certificateauthority",
            name="name",
            field=models.CharField(help_text="A human-readable name", max_length=128, unique=True),
        ),
        migrations.AlterField(
            model_name="certificateauthority",
            name="ocsp_url",
            field=models.URLField(
                blank=True,
                default="",
                help_text="URL of a OCSP responser for the CA.",
                verbose_name="OCSP responder URL",
            ),
            preserve_default=False,
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""django-ca custom URL converters.

.. seealso:: https://docs.djangoproject.com/en/dev/topics/http/urls/
"""
# pylint: disable=missing-function-docstring; All functions are given by Django

from django.urls.converters import SlugConverter

from django_ca.utils import sanitize_serial


class HexConverter:
    """Converter that accepts colon-separated hex values."""

    regex = "[0-9A-F:]+"

    def to_python(self, value: str) -> str:
        return value

    def to_url(self, value: str) -> str:
        return value


class SerialConverter(HexConverter):
    """Extends base to call :py:func:`~django-ca.utils.sanitize_serial` for the value."""

    def to_python(self, value: str) -> str:
        return sanitize_serial(value)


class Base64Converter:
    """Converter that accepts Base64 encoded data."""

    regex = "[a-zA-Z0-9=+/]+"

    def to_python(self, value: str) -> str:
        return value

    def to_url(self, value: str) -> str:
        return value


class AcmeSlugConverter(SlugConverter):
    """ACME slugs consist of alphanumeric characters only."""

    regex = "[a-zA-Z0-9]+"
